"""Provides the ContextHandler class for managing global MadsciContext throughout a MADSci system component."""

import asyncio
import contextlib
import contextvars
import functools
import inspect
from collections.abc import Callable, Generator
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    TypeVar,
    Union,
    overload,
)

from madsci.common.types.context_types import MadsciContext
from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from madsci.client.event_client import EventClient
    from madsci.common.types.event_types import EventClientContext

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")

__all__ = [
    "GlobalMadsciContext",
    "_event_client_context",
    "event_client_class",
    "event_client_context",
    "get_current_madsci_context",
    "get_event_client",
    "get_event_client_context",
    "has_event_client_context",
    "has_madsci_context",
    "madsci_context",
    "madsci_context_class",
    "set_current_madsci_context",
    "with_event_client",
    "with_madsci_context",
]


class GlobalMadsciContext:
    """
    A global MadsciContext object for the application.

    This singleton can be accessed from anywhere in the codebase, but should
    not be modified directly. Instead, use the madsci_context context manager
    to temporarily override values in the context.
    """

    _context: Optional[MadsciContext] = None

    @classmethod
    def get_context(cls) -> MadsciContext:
        """
        Get the global context, creating it lazily if needed.

        Returns:
            The global MadsciContext instance.
        """
        if cls._context is None:
            cls._context = MadsciContext()
        return cls._context

    @classmethod
    def set_context(cls, context: MadsciContext) -> None:
        """
        Set the global context.

        Args:
            context: The MadsciContext instance to set as global.
        """
        cls._context = context


_current_madsci_context: contextvars.ContextVar[Optional[MadsciContext]] = (
    contextvars.ContextVar(
        "current_madsci_context",
        default=None,
    )
)


@contextlib.contextmanager
def madsci_context(**overrides: Any) -> Generator[MadsciContext, None, None]:
    """Updates the current MadsciContext (as returned by get_current_madsci_context) with the provided overrides."""
    prev_context = _current_madsci_context.get()
    if prev_context is None:
        prev_context = GlobalMadsciContext.get_context()
    context = prev_context.model_copy()
    for k, v in overrides.items():
        setattr(context, k, v)
    token = _current_madsci_context.set(context)
    try:
        yield _current_madsci_context.get()  # type: ignore[misc]
    finally:
        _current_madsci_context.reset(token)


def get_current_madsci_context() -> MadsciContext:
    """Returns the current MadsciContext object."""
    context = _current_madsci_context.get()
    if context is None:
        context = GlobalMadsciContext.get_context()
        _current_madsci_context.set(context)
    return context


def set_current_madsci_context(context: MadsciContext) -> None:
    """Sets the current MadsciContext object."""
    _current_madsci_context.set(context)


def has_madsci_context() -> bool:
    """
    Check if a MadsciContext has been explicitly set (not just using global default).

    Returns:
        True if a context has been explicitly set, False otherwise.
    """
    return _current_madsci_context.get() is not None


# =============================================================================
# MadsciContext Decorators
# =============================================================================


@overload
def with_madsci_context(
    func: Callable[P, R],
) -> Callable[P, R]: ...


@overload
def with_madsci_context(
    *,
    lab_server_url: Optional[str] = None,
    event_server_url: Optional[str] = None,
    experiment_server_url: Optional[str] = None,
    data_server_url: Optional[str] = None,
    resource_server_url: Optional[str] = None,
    workcell_server_url: Optional[str] = None,
    location_server_url: Optional[str] = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...


def with_madsci_context(
    func: Optional[Callable[P, R]] = None,
    *,
    lab_server_url: Optional[str] = None,
    event_server_url: Optional[str] = None,
    experiment_server_url: Optional[str] = None,
    data_server_url: Optional[str] = None,
    resource_server_url: Optional[str] = None,
    workcell_server_url: Optional[str] = None,
    location_server_url: Optional[str] = None,
) -> Union[Callable[P, R], Callable[[Callable[P, R]], Callable[P, R]]]:
    """
    Decorator that establishes a MadsciContext for a function.

    This decorator wraps a function (sync or async) so that all code executed
    within it has access to the specified MadsciContext configuration via
    get_current_madsci_context(). The decorated function can optionally receive
    the MadsciContext as a keyword argument named 'madsci_ctx'.

    Can be used with or without arguments:
        @with_madsci_context
        def my_function(): ...

        @with_madsci_context(event_server_url="http://localhost:8001")
        def my_workflow(): ...

    Args:
        func: The function to wrap (when used without parentheses).
        lab_server_url: Lab server URL to set in context.
        event_server_url: Event server URL to set in context.
        experiment_server_url: Experiment server URL to set in context.
        data_server_url: Data server URL to set in context.
        resource_server_url: Resource server URL to set in context.
        workcell_server_url: Workcell server URL to set in context.
        location_server_url: Location server URL to set in context.

    Returns:
        The decorated function that runs within a MadsciContext.

    Example:
        @with_madsci_context(event_server_url="http://localhost:8001")
        def log_events():
            ctx = get_current_madsci_context()
            print(f"Logging to: {ctx.event_server_url}")

        @with_madsci_context(lab_server_url="http://lab:8000")
        def connect_to_lab(madsci_ctx: MadsciContext = None):
            # madsci_ctx is automatically injected
            print(f"Connecting to: {madsci_ctx.lab_server_url}")

        @with_madsci_context(data_server_url="http://data:8004")
        async def async_data_op():
            ctx = get_current_madsci_context()
            await upload_data(ctx.data_server_url)
    """
    # Collect non-None overrides
    overrides = {
        k: v
        for k, v in {
            "lab_server_url": lab_server_url,
            "event_server_url": event_server_url,
            "experiment_server_url": experiment_server_url,
            "data_server_url": data_server_url,
            "resource_server_url": resource_server_url,
            "workcell_server_url": workcell_server_url,
            "location_server_url": location_server_url,
        }.items()
        if v is not None
    }

    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        # Check if the function accepts 'madsci_ctx' parameter
        sig = inspect.signature(fn)
        accepts_madsci_ctx = "madsci_ctx" in sig.parameters

        if asyncio.iscoroutinefunction(fn):

            @functools.wraps(fn)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                with madsci_context(**overrides) as ctx:
                    if accepts_madsci_ctx and "madsci_ctx" not in kwargs:
                        kwargs["madsci_ctx"] = ctx  # type: ignore[assignment]
                    return await fn(*args, **kwargs)

            return async_wrapper  # type: ignore[return-value]

        @functools.wraps(fn)
        def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            with madsci_context(**overrides) as ctx:
                if accepts_madsci_ctx and "madsci_ctx" not in kwargs:
                    kwargs["madsci_ctx"] = ctx  # type: ignore[assignment]
                return fn(*args, **kwargs)

        return sync_wrapper  # type: ignore[return-value]

    # Handle both @with_madsci_context and @with_madsci_context(...) syntax
    if func is not None:
        return decorator(func)
    return decorator


def madsci_context_class(
    lab_server_url: Optional[str] = None,
    event_server_url: Optional[str] = None,
    experiment_server_url: Optional[str] = None,
    data_server_url: Optional[str] = None,
    resource_server_url: Optional[str] = None,
    workcell_server_url: Optional[str] = None,
    location_server_url: Optional[str] = None,
) -> Callable[[type], type]:
    """
    Class decorator that establishes a MadsciContext for method calls.

    This decorator modifies a class so that all public methods run within a
    MadsciContext. The class gains a 'madsci_context' property that returns
    the current context.

    Args:
        lab_server_url: Lab server URL to set in context.
        event_server_url: Event server URL to set in context.
        experiment_server_url: Experiment server URL to set in context.
        data_server_url: Data server URL to set in context.
        resource_server_url: Resource server URL to set in context.
        workcell_server_url: Workcell server URL to set in context.
        location_server_url: Location server URL to set in context.

    Returns:
        A class decorator.

    Example:
        @madsci_context_class(event_server_url="http://localhost:8001")
        class EventLogger:
            def log_event(self, message):
                # self.madsci_context is available
                print(f"Logging to: {self.madsci_context.event_server_url}")

        @madsci_context_class(lab_server_url="http://lab:8000")
        class LabConnector:
            def get_context_overrides(self) -> dict:
                # Add instance-specific context overrides
                return {"workcell_server_url": self.workcell_url}

            def connect(self):
                # Uses both class-level and instance-level context
                print(f"Lab: {self.madsci_context.lab_server_url}")
    """
    # Collect non-None overrides
    class_overrides = {
        k: v
        for k, v in {
            "lab_server_url": lab_server_url,
            "event_server_url": event_server_url,
            "experiment_server_url": experiment_server_url,
            "data_server_url": data_server_url,
            "resource_server_url": resource_server_url,
            "workcell_server_url": workcell_server_url,
            "location_server_url": location_server_url,
        }.items()
        if v is not None
    }

    def class_decorator(cls: type) -> type:
        # Store overrides on the class
        cls._madsci_context_overrides = class_overrides  # type: ignore[attr-defined]

        # Add madsci_context property if needed
        _add_madsci_context_property(cls)

        # Wrap public methods
        _wrap_madsci_context_methods(cls, class_overrides)

        return cls

    return class_decorator


def _add_madsci_context_property(cls: type) -> None:
    """Add a madsci_context property to a class if it doesn't already have one."""
    if hasattr(cls, "madsci_context") and isinstance(
        getattr(cls, "madsci_context", None), property
    ):
        return

    @property
    def madsci_context_property(self: Any) -> MadsciContext:  # noqa: ARG001
        """Get the current MadsciContext."""
        return get_current_madsci_context()

    cls.madsci_context = madsci_context_property  # type: ignore[attr-defined]


def _should_wrap_madsci_context_method(cls: type, attr_name: str) -> bool:
    """Determine if a class attribute should be wrapped with madsci context."""
    # Skip private/magic methods
    if attr_name.startswith("_"):
        return False

    # Skip methods that could cause recursion with stacked decorators
    if attr_name in _DECORATOR_EXCLUDED_METHODS:
        return False

    attr = getattr(cls, attr_name, None)

    # Only wrap callable attributes
    if not callable(attr):
        return False

    # Skip class attributes that are types (classes) - they are callable but not methods
    # This prevents wrapping class attributes like DEFINITION_CLASS, SETTINGS_CLASS, etc.
    if isinstance(attr, type):
        return False

    # Skip classmethods, staticmethods, and properties
    static_attr = inspect.getattr_static(cls, attr_name)
    return not isinstance(static_attr, (classmethod, staticmethod, property))


def _wrap_madsci_context_methods(cls: type, class_overrides: dict[str, Any]) -> None:
    """Wrap all public methods of a class with madsci context."""
    for attr_name in dir(cls):
        if not _should_wrap_madsci_context_method(cls, attr_name):
            continue

        attr = getattr(cls, attr_name)
        wrapped = _wrap_madsci_context_method(attr, class_overrides)
        setattr(cls, attr_name, wrapped)


def _wrap_madsci_context_method(
    method: Callable[..., R],
    class_overrides: dict[str, Any],
) -> Callable[..., R]:
    """Wrap a method to run within a madsci context."""
    if asyncio.iscoroutinefunction(method):

        @functools.wraps(method)
        async def async_method_wrapper(self: Any, *args: Any, **kwargs: Any) -> R:
            # Get instance-specific overrides if available
            overrides = dict(class_overrides)
            if hasattr(self, "get_context_overrides"):
                overrides.update(self.get_context_overrides())

            with madsci_context(**overrides):
                return await method(self, *args, **kwargs)

        return async_method_wrapper  # type: ignore[return-value]

    @functools.wraps(method)
    def sync_method_wrapper(self: Any, *args: Any, **kwargs: Any) -> R:
        # Get instance-specific overrides if available
        overrides = dict(class_overrides)
        if hasattr(self, "get_context_overrides"):
            overrides.update(self.get_context_overrides())

        with madsci_context(**overrides):
            return method(self, *args, **kwargs)

    return sync_method_wrapper  # type: ignore[return-value]


# =============================================================================
# EventClient Context Management
# =============================================================================

_event_client_context: contextvars.ContextVar[Optional["EventClientContext"]] = (
    contextvars.ContextVar(
        "event_client_context",
        default=None,
    )
)


def get_event_client(
    name: Optional[str] = None,
    create_if_missing: bool = True,
    **context_kwargs: Any,
) -> "EventClient":
    """
    Get the current EventClient from context, or create one if none exists.

    This is the primary way to obtain an EventClient in MADSci code.
    It enables automatic context propagation and hierarchical logging.

    Args:
        name: Optional name override. If provided without existing context,
              uses this as the client name. Ignored if context exists.
        create_if_missing: If True (default), create a new EventClient when
                          no context exists. If False, raises RuntimeError.
        **context_kwargs: Additional context to bind to the returned client.
                         If context exists, returns a bound child client.
                         If creating new, these become initial bound context.

    Returns:
        An EventClient instance, potentially with inherited context.

    Raises:
        RuntimeError: If create_if_missing=False and no context exists.

    Example:
        # In a component that should inherit context:
        logger = get_event_client()
        logger.info("Processing...")

        # In a component that wants to add context:
        logger = get_event_client(node_id="node-123")
        logger.info("Node action")  # Includes node_id
    """
    ctx = _event_client_context.get()

    if ctx is not None:
        # We have an existing context - use it or extend it
        if context_kwargs:
            # Caller wants to add context - return a bound child
            return ctx.client.bind(**context_kwargs)
        return ctx.client

    # No context exists
    if not create_if_missing:
        raise RuntimeError(
            "No EventClient context available. "
            "Wrap your code in an event_client_context() block or use create_if_missing=True."
        )

    # Create a new root EventClient
    # Use stack inspection as fallback for name
    if name is None:
        name = _infer_caller_name()

    from madsci.client.event_client import EventClient  # noqa: PLC0415

    client = EventClient(name=name)

    # Bind any provided context
    if context_kwargs:
        client = client.bind(**context_kwargs)

    return client


def _infer_caller_name() -> str:
    """
    Infer a reasonable name from the call stack.

    Walks up the stack to find the first frame that's not in the
    context module, and uses its module name.

    Uses inspect.currentframe() and f_back traversal instead of
    inspect.stack() for better performance (avoids building full stack list).

    Returns:
        Inferred module name, or "madsci" as fallback.
    """
    frame = inspect.currentframe()
    # Skip this function's own frame
    if frame is not None:
        frame = frame.f_back

    # Walk back through frames without constructing the full stack
    try:
        depth = 0
        while frame is not None and depth < 50:
            module = frame.f_globals.get("__name__", "")
            # Skip context module frames
            if not module.startswith("madsci.common.context"):
                return module
            frame = frame.f_back
            depth += 1
    finally:
        # Avoid reference cycles involving frame objects
        del frame

    return "madsci"


@contextlib.contextmanager
def event_client_context(
    name: Optional[str] = None,
    client: Optional["EventClient"] = None,
    inherit: bool = True,
    **context_metadata: Any,
) -> Generator["EventClient", None, None]:
    """
    Establish or extend an EventClient context.

    All code executed within this context manager will have access to
    the EventClient via get_event_client(), with accumulated context.

    Args:
        name: Name for this context level (e.g., "workflow", "node").
              Added to the hierarchy.
        client: Explicit EventClient to use. If None, inherits from
               parent context or creates a new one.
        inherit: If True (default), inherit and extend parent context.
                If False, create a fresh context.
        **context_metadata: Additional context to bind to all log messages
                           (e.g., experiment_id, workflow_id).

    Yields:
        The EventClient for this context.

    Example:
        # Establish root context in an experiment
        with event_client_context(name="experiment", experiment_id="exp-123") as logger:
            logger.info("Starting experiment")

            # Nested context for workflow
            with event_client_context(name="workflow", workflow_id="wf-456") as wf_logger:
                wf_logger.info("Running workflow")
                # Log includes both experiment_id and workflow_id
    """
    from madsci.client.event_client import EventClient  # noqa: PLC0415
    from madsci.common.types.event_types import EventClientContext  # noqa: PLC0415

    parent_ctx = _event_client_context.get()

    if client is not None:
        # Explicit client provided - use it
        new_client = client.bind(**context_metadata) if context_metadata else client
        hierarchy = [name] if name else []
        new_ctx = EventClientContext(
            client=new_client,
            hierarchy=hierarchy,
            metadata=dict(context_metadata),
        )
    elif parent_ctx is not None and inherit:
        # Inherit from parent context
        # Use provided name, or generate one from caller if not provided
        effective_name = name or _infer_caller_name().split(".")[-1]
        new_ctx = parent_ctx.child(effective_name, **context_metadata)
    else:
        # Create new root context
        effective_name = name or _infer_caller_name()
        new_client = EventClient(name=effective_name)
        if context_metadata:
            new_client = new_client.bind(**context_metadata)
        # Use effective_name for hierarchy to ensure consistency
        new_ctx = EventClientContext(
            client=new_client,
            hierarchy=[effective_name] if effective_name else [],
            metadata=dict(context_metadata),
        )

    token = _event_client_context.set(new_ctx)
    try:
        yield new_ctx.client
    finally:
        _event_client_context.reset(token)


def get_event_client_context() -> Optional["EventClientContext"]:
    """
    Get the current EventClientContext, if any.

    Returns:
        The current EventClientContext, or None if no context is active.

    Example:
        ctx = get_event_client_context()
        if ctx:
            print(f"Current hierarchy: {ctx.name}")
            print(f"Metadata: {ctx.metadata}")
    """
    return _event_client_context.get()


def has_event_client_context() -> bool:
    """
    Check if an EventClient context is currently active.

    Returns:
        True if a context is active, False otherwise.

    Example:
        if has_event_client_context():
            logger = get_event_client()  # Uses context
        else:
            logger = EventClient(name="standalone")  # Create new
    """
    return _event_client_context.get() is not None


# =============================================================================
# EventClient Context Decorators
# =============================================================================


@overload
def with_event_client(
    func: Callable[P, R],
) -> Callable[P, R]: ...


@overload
def with_event_client(
    *,
    name: Optional[str] = None,
    inherit: bool = True,
    **context_metadata: Any,
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...


def with_event_client(
    func: Optional[Callable[P, R]] = None,
    *,
    name: Optional[str] = None,
    inherit: bool = True,
    **context_metadata: Any,
) -> Union[Callable[P, R], Callable[[Callable[P, R]], Callable[P, R]]]:
    """
    Decorator that establishes an EventClient context for a function.

    This decorator wraps a function (sync or async) so that all code executed
    within it has access to an EventClient via get_event_client(), with
    accumulated context. The decorated function can optionally receive the
    EventClient as a keyword argument named 'event_client'.

    Can be used with or without arguments:
        @with_event_client
        def my_function(): ...

        @with_event_client(name="my_workflow", workflow_id="wf-123")
        def my_workflow(): ...

    Args:
        func: The function to wrap (when used without parentheses).
        name: Name for this context level. If not provided, uses the
              function's qualified name.
        inherit: If True (default), inherit and extend parent context.
                If False, create a fresh context.
        **context_metadata: Additional context to bind to all log messages
                           within this function (e.g., experiment_id, node_id).

    Returns:
        The decorated function that runs within an EventClient context.

    Example:
        @with_event_client
        def process_data():
            logger = get_event_client()
            logger.info("Processing data...")

        @with_event_client(name="experiment", experiment_id="exp-123")
        def run_experiment(event_client: EventClient = None):
            # event_client is automatically injected
            event_client.info("Running experiment")

        @with_event_client(name="async_task")
        async def async_operation():
            logger = get_event_client()
            await some_async_work()
            logger.info("Done")
    """

    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        # Determine the context name
        context_name = name or fn.__qualname__

        # Check if the function accepts 'event_client' parameter
        sig = inspect.signature(fn)
        accepts_event_client = "event_client" in sig.parameters

        if asyncio.iscoroutinefunction(fn):
            # Async function wrapper
            @functools.wraps(fn)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                with event_client_context(
                    name=context_name,
                    inherit=inherit,
                    **context_metadata,
                ) as client:
                    if accepts_event_client and "event_client" not in kwargs:
                        kwargs["event_client"] = client  # type: ignore[assignment]
                    return await fn(*args, **kwargs)

            return async_wrapper  # type: ignore[return-value]

        # Sync function wrapper
        @functools.wraps(fn)
        def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            with event_client_context(
                name=context_name,
                inherit=inherit,
                **context_metadata,
            ) as client:
                if accepts_event_client and "event_client" not in kwargs:
                    kwargs["event_client"] = client  # type: ignore[assignment]
                return fn(*args, **kwargs)

        return sync_wrapper  # type: ignore[return-value]

    # Handle both @with_event_client and @with_event_client(...) syntax
    if func is not None:
        return decorator(func)
    return decorator


def event_client_class(
    name: Optional[str] = None,
    inherit: bool = True,
    **context_metadata: Any,
) -> Callable[[type[T]], type[T]]:
    """
    Class decorator that establishes an EventClient context for method calls.

    This decorator modifies a class so that specified methods (or all public
    methods if none specified) run within an EventClient context. The class
    gains an 'event_client' property that returns the current context's client.

    Args:
        name: Base name for the context. Defaults to the class name.
              Method contexts will be named "{class_name}.{method_name}".
        inherit: If True (default), inherit and extend parent context.
                If False, create a fresh context for each method call.
        **context_metadata: Additional context to bind to all log messages
                           within this class (e.g., component_type="processor").

    Returns:
        A class decorator.

    Example:
        @event_client_class(component_type="data_processor")
        class DataProcessor:
            def process(self, data):
                # self.event_client is available
                self.event_client.info("Processing", data_size=len(data))
                return self._transform(data)

            def _transform(self, data):  # Private methods not wrapped
                return data.upper()

        @event_client_class(name="CustomName", wrap_methods=["run", "execute"])
        class Workflow:
            def run(self):
                self.event_client.info("Running workflow")

            def execute(self):
                self.event_client.info("Executing")

            def helper(self):  # Not wrapped
                pass
    """

    def class_decorator(cls: type[T]) -> type[T]:
        class_name = name or cls.__name__

        # Store context settings on the class
        cls._event_client_context_name = class_name  # type: ignore[attr-defined]
        cls._event_client_context_inherit = inherit  # type: ignore[attr-defined]
        cls._event_client_context_metadata = context_metadata  # type: ignore[attr-defined]

        # Add event_client property if needed
        _add_event_client_property(cls)

        # Wrap public methods
        _wrap_public_methods(cls, class_name, inherit, context_metadata)

        return cls

    return class_decorator


def _add_event_client_property(cls: type) -> None:
    """Add an event_client property to a class if it doesn't already have one."""
    if hasattr(cls, "event_client") and isinstance(
        getattr(cls, "event_client", None), property
    ):
        return

    @property
    def event_client_property(self: Any) -> "EventClient":
        """Get the EventClient for this instance from the current context."""
        instance_metadata = {
            **self._event_client_context_metadata,
        }
        # Add instance-specific context if available
        if hasattr(self, "get_event_context"):
            instance_metadata.update(self.get_event_context())
        return get_event_client(**instance_metadata)

    cls.event_client = event_client_property  # type: ignore[attr-defined]


# Methods that should never be wrapped by class decorators to avoid recursion
# when multiple decorators are stacked
_DECORATOR_EXCLUDED_METHODS = frozenset(
    {
        # EventClient context
        "get_event_context",
        # Ownership context
        "get_ownership_overrides",
        # MadsciContext
        "get_context_overrides",
        # OTEL tracing
        "get_span_attributes",
    }
)


def _should_wrap_method(cls: type, attr_name: str) -> bool:
    """Determine if a class attribute should be wrapped with event client context."""
    # Skip private/magic methods
    if attr_name.startswith("_"):
        return False

    # Skip methods that could cause recursion with stacked decorators
    if attr_name in _DECORATOR_EXCLUDED_METHODS:
        return False

    attr = getattr(cls, attr_name, None)

    # Only wrap callable attributes
    if not callable(attr):
        return False

    # Skip class attributes that are types (classes) - they are callable but not methods
    # This prevents wrapping class attributes like DEFINITION_CLASS, SETTINGS_CLASS, etc.
    if isinstance(attr, type):
        return False

    # Skip classmethods, staticmethods, and properties
    static_attr = inspect.getattr_static(cls, attr_name)
    return not isinstance(static_attr, (classmethod, staticmethod, property))


def _wrap_public_methods(
    cls: type,
    class_name: str,
    inherit: bool,
    context_metadata: dict[str, Any],
) -> None:
    """Wrap all public methods of a class with event client context."""
    for attr_name in dir(cls):
        if not _should_wrap_method(cls, attr_name):
            continue

        attr = getattr(cls, attr_name)
        method_context_name = f"{class_name}.{attr_name}"
        wrapped = _wrap_method(attr, method_context_name, inherit, context_metadata)
        setattr(cls, attr_name, wrapped)


def _wrap_method(
    method: Callable[..., R],
    context_name: str,
    inherit: bool,
    context_metadata: dict[str, Any],
) -> Callable[..., R]:
    """
    Wrap a method to run within an EventClient context.

    Args:
        method: The method to wrap.
        context_name: Name for the context.
        inherit: Whether to inherit parent context.
        context_metadata: Additional context metadata.

    Returns:
        The wrapped method.
    """
    if asyncio.iscoroutinefunction(method):

        @functools.wraps(method)
        async def async_method_wrapper(self: Any, *args: Any, **kwargs: Any) -> R:
            # Get instance-specific context if available
            instance_metadata = dict(context_metadata)
            if hasattr(self, "get_event_context"):
                instance_metadata.update(self.get_event_context())

            with event_client_context(
                name=context_name,
                inherit=inherit,
                **instance_metadata,
            ):
                return await method(self, *args, **kwargs)

        return async_method_wrapper  # type: ignore[return-value]

    @functools.wraps(method)
    def sync_method_wrapper(self: Any, *args: Any, **kwargs: Any) -> R:
        # Get instance-specific context if available
        instance_metadata = dict(context_metadata)
        if hasattr(self, "get_event_context"):
            instance_metadata.update(self.get_event_context())

        with event_client_context(
            name=context_name,
            inherit=inherit,
            **instance_metadata,
        ):
            return method(self, *args, **kwargs)

    return sync_method_wrapper  # type: ignore[return-value]



"""Test MongoDB backup CLI interface."""

import json
import tempfile
from datetime import datetime
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner
from madsci.common.backup_tools.base_backup import BackupInfo
from madsci.common.backup_tools.mongo_cli import mongodb_backup


class TestMongoDBBackupCLI:
    """Test MongoDB backup CLI interface."""

    @pytest.fixture
    def temp_backup_dir(self):
        """Create temporary backup directory for testing."""
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)

    @pytest.fixture
    def cli_runner(self):
        """Create Click CLI runner."""
        return CliRunner()

    @pytest.fixture
    def mock_backup_tool(self, temp_backup_dir):
        """Create mock backup tool for CLI testing."""
        mock_tool = Mock()
        mock_tool.create_backup.return_value = temp_backup_dir / "test_backup"
        mock_tool.list_available_backups.return_value = []
        mock_tool.validate_backup_integrity.return_value = True
        mock_tool.restore_from_backup.return_value = None
        mock_tool.delete_backup.return_value = None
        return mock_tool

    def test_cli_create_backup(self, cli_runner, temp_backup_dir, mock_backup_tool):
        """Test CLI backup creation command."""

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool
            mock_backup_tool.create_backup.return_value = Path(
                str(temp_backup_dir / "test_backup")
            )

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "--backup-dir",
                    str(temp_backup_dir),
                    "create",
                ],
            )

            assert result.exit_code == 0
            assert "Backup created successfully" in result.output
            mock_backup_tool.create_backup.assert_called_once_with(None)

    def test_cli_create_backup_with_name_suffix(self, cli_runner, mock_backup_tool):
        """Test CLI backup creation with custom name suffix."""

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "create",
                    "--name-suffix",
                    "pre_migration",
                ],
            )

            assert result.exit_code == 0
            mock_backup_tool.create_backup.assert_called_once_with("pre_migration")

    def test_cli_create_backup_no_validation(self, cli_runner, mock_backup_tool):
        """Test CLI backup creation with validation disabled."""

        with (
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
            ) as mock_tool_class,
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupSettings"
            ) as mock_settings,
        ):
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "create",
                    "--no-validate",
                ],
            )

            assert result.exit_code == 0
            # Verify settings were created with validation disabled
            mock_settings.assert_called_once()
            call_kwargs = mock_settings.call_args.kwargs
            assert call_kwargs["validate_integrity"] is False

    def test_cli_create_backup_specific_collections(self, cli_runner, mock_backup_tool):
        """Test CLI backup creation for specific collections."""

        with (
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
            ) as mock_tool_class,
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupSettings"
            ) as mock_settings,
        ):
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "create",
                    "--collections",
                    "collection1,collection2",
                ],
            )

            assert result.exit_code == 0
            # Verify collections were passed to settings
            mock_settings.assert_called_once()
            call_kwargs = mock_settings.call_args.kwargs
            assert call_kwargs["collections"] == ["collection1", "collection2"]

    def test_cli_list_backups(self, cli_runner, temp_backup_dir, mock_backup_tool):
        """Test CLI backup listing command."""
        # Create mock backup info (use larger size to see in MB)
        mock_backup_info = BackupInfo(
            backup_path=temp_backup_dir / "test_backup",
            created_at=datetime(2024, 1, 1, 12, 0, 0),
            database_version="1.0.0",
            backup_size=2097152,  # 2 MB in bytes
            checksum="test_checksum",
            backup_type="mongodb",
            is_valid=True,
        )

        mock_backup_tool.list_available_backups.return_value = [mock_backup_info]

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "list",
                ],
            )

            assert result.exit_code == 0
            assert "test_backup" in result.output
            assert "2024-01-01 12:00:00" in result.output
            assert "2.0 MB" in result.output  # 2097152 bytes displayed as 2.0 MB
            mock_backup_tool.list_available_backups.assert_called_once()

    def test_cli_list_backups_empty(self, cli_runner, mock_backup_tool):
        """Test CLI backup listing with no backups."""

        mock_backup_tool.list_available_backups.return_value = []

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "list",
                ],
            )

            assert result.exit_code == 0
            assert "No backups found" in result.output

    def test_cli_restore_backup(self, cli_runner, temp_backup_dir, mock_backup_tool):
        """Test CLI backup restoration command."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "restore",
                    str(backup_path),
                ],
            )

            assert result.exit_code == 0
            assert "Restore completed successfully" in result.output
            mock_backup_tool.restore_from_backup.assert_called_once_with(
                backup_path, None
            )

    def test_cli_restore_to_different_database(
        self, cli_runner, temp_backup_dir, mock_backup_tool
    ):
        """Test CLI restore to different target database."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "restore",
                    str(backup_path),
                    "--target-database",
                    "target_db",
                ],
            )

            assert result.exit_code == 0
            mock_backup_tool.restore_from_backup.assert_called_once_with(
                backup_path, "target_db"
            )

    def test_cli_validate_backup(self, cli_runner, temp_backup_dir, mock_backup_tool):
        """Test CLI backup validation command."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool
            mock_backup_tool.validate_backup_integrity.return_value = True

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "validate",
                    str(backup_path),
                ],
            )

            assert result.exit_code == 0
            assert "Backup validation successful" in result.output
            mock_backup_tool.validate_backup_integrity.assert_called_once_with(
                backup_path
            )

    def test_cli_validate_backup_failure(
        self, cli_runner, temp_backup_dir, mock_backup_tool
    ):
        """Test CLI backup validation failure."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool
            mock_backup_tool.validate_backup_integrity.return_value = False

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "validate",
                    str(backup_path),
                ],
            )

            assert result.exit_code == 1
            assert "Backup validation failed" in result.output

    def test_cli_delete_backup(self, cli_runner, temp_backup_dir, mock_backup_tool):
        """Test CLI backup deletion command."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            # Provide 'y' input to confirm deletion
            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "delete",
                    str(backup_path),
                ],
                input="y\n",
            )

            assert result.exit_code == 0
            assert "Backup deleted successfully" in result.output
            mock_backup_tool.delete_backup.assert_called_once_with(backup_path)

    def test_cli_delete_backup_cancelled(
        self, cli_runner, temp_backup_dir, mock_backup_tool
    ):
        """Test CLI backup deletion cancelled by user."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            # Provide 'n' input to cancel deletion
            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "delete",
                    str(backup_path),
                ],
                input="n\n",
            )

            assert result.exit_code == 0
            assert "Deletion cancelled" in result.output
            mock_backup_tool.delete_backup.assert_not_called()

    def test_cli_delete_backup_force(
        self, cli_runner, temp_backup_dir, mock_backup_tool
    ):
        """Test CLI backup deletion with force flag."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "delete",
                    str(backup_path),
                    "--force",
                ],
            )

            assert result.exit_code == 0
            assert "Backup deleted successfully" in result.output
            mock_backup_tool.delete_backup.assert_called_once_with(backup_path)

    def test_cli_error_handling_create_failure(self, cli_runner, mock_backup_tool):
        """Test CLI error handling for backup creation failure."""

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool
            mock_backup_tool.create_backup.side_effect = RuntimeError("Backup failed")

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "create",
                ],
            )

            assert result.exit_code == 1
            assert "Backup failed" in result.output

    def test_cli_error_handling_restore_failure(
        self, cli_runner, temp_backup_dir, mock_backup_tool
    ):
        """Test CLI error handling for restore failure."""

        backup_path = temp_backup_dir / "test_backup"
        backup_path.mkdir()

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool
            mock_backup_tool.restore_from_backup.side_effect = RuntimeError(
                "Restore failed"
            )

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "restore",
                    str(backup_path),
                ],
            )

            assert result.exit_code == 1
            assert "Restore failed" in result.output

    def test_cli_invalid_backup_path(self, cli_runner, mock_backup_tool):
        """Test CLI with non-existent backup path."""

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "restore",
                    "/non/existent/path",
                ],
            )

            assert result.exit_code == 1
            assert "Backup path does not exist" in result.output

    def test_cli_configuration_file_support(
        self, cli_runner, temp_backup_dir, mock_backup_tool
    ):
        """Test CLI supports configuration files."""

        # Create config file
        config_file = temp_backup_dir / "config.json"
        config_data = {
            "mongo_db_url": "mongodb://localhost:27017/",
            "database": "test_db",
            "backup_dir": str(temp_backup_dir),
            "max_backups": 5,
        }
        config_file.write_text(json.dumps(config_data))

        with (
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
            ) as mock_tool_class,
            patch("madsci.common.backup_tools.mongo_cli.MongoDBBackupSettings"),
        ):
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup, ["--config-file", str(config_file), "create"]
            )

            assert result.exit_code == 0

    def test_cli_help_commands(self, cli_runner):
        """Test CLI help commands work correctly."""

        # Test main help
        result = cli_runner.invoke(mongodb_backup, ["--help"])
        assert result.exit_code == 0
        assert "MongoDB backup management commands" in result.output

        # Test create help (provide required parent options)
        result = cli_runner.invoke(
            mongodb_backup,
            [
                "--mongo-url",
                "mongodb://localhost:27017/",
                "--database",
                "test_db",
                "create",
                "--help",
            ],
        )
        assert result.exit_code == 0
        assert "Create a new MongoDB backup" in result.output

        # Test restore help (provide required parent options)
        result = cli_runner.invoke(
            mongodb_backup,
            [
                "--mongo-url",
                "mongodb://localhost:27017/",
                "--database",
                "test_db",
                "restore",
                "--help",
            ],
        )
        assert result.exit_code == 0
        assert "Restore from MongoDB backup" in result.output

    def test_cli_missing_required_options(self, cli_runner):
        """Test CLI with missing required options."""

        # Test create without required options
        result = cli_runner.invoke(mongodb_backup, ["create"])
        assert result.exit_code != 0
        assert "MongoDB URL is required" in result.output

    def test_cli_verbose_output(self, cli_runner, mock_backup_tool):
        """Test CLI verbose output option."""

        with patch(
            "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
        ) as mock_tool_class:
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "--verbose",
                    "create",
                ],
            )

            assert result.exit_code == 0

    def test_cli_compression_option(self, cli_runner, mock_backup_tool):
        """Test CLI compression option."""

        with (
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupTool"
            ) as mock_tool_class,
            patch(
                "madsci.common.backup_tools.mongo_cli.MongoDBBackupSettings"
            ) as mock_settings,
        ):
            mock_tool_class.return_value = mock_backup_tool

            result = cli_runner.invoke(
                mongodb_backup,
                [
                    "--mongo-url",
                    "mongodb://localhost:27017/",
                    "--database",
                    "test_db",
                    "create",
                    "--no-compression",
                ],
            )

            assert result.exit_code == 0
            # Verify compression was disabled
            mock_settings.assert_called_once()
            call_kwargs = mock_settings.call_args.kwargs
            assert call_kwargs["compression"] is False



"""Tests for the deprecation utilities."""

import warnings
from pathlib import Path
from typing import Optional

from madsci.common.deprecation import (
    DEPRECATED_IN,
    REMOVAL_IN,
    MadsciDeprecationWarning,
    deprecated,
    deprecated_parameter,
    emit_definition_deprecation_warning,
)


class TestEmitDefinitionDeprecationWarning:
    """Tests for the emit_definition_deprecation_warning function."""

    def test_emits_warning(self, tmp_path: Path) -> None:
        """Test that the function emits a deprecation warning."""
        def_path = tmp_path / "test.manager.yaml"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            emit_definition_deprecation_warning(def_path, "manager")

            assert len(w) == 1
            assert issubclass(w[0].category, MadsciDeprecationWarning)
            assert "manager" in str(w[0].message)
            assert str(def_path) in str(w[0].message)

    def test_includes_version_info(self, tmp_path: Path) -> None:
        """Test that the warning includes version information."""
        def_path = tmp_path / "test.manager.yaml"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            emit_definition_deprecation_warning(def_path, "manager")

            message = str(w[0].message)
            assert DEPRECATED_IN in message
            assert REMOVAL_IN in message

    def test_includes_migration_command(self, tmp_path: Path) -> None:
        """Test that the warning includes the migration command."""
        def_path = tmp_path / "test.manager.yaml"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            emit_definition_deprecation_warning(
                def_path,
                "manager",
                migration_command="madsci migrate convert test.manager.yaml",
            )

            message = str(w[0].message)
            assert "madsci migrate convert test.manager.yaml" in message

    def test_default_migration_hint(self, tmp_path: Path) -> None:
        """Test default migration hint when no command specified."""
        def_path = tmp_path / "test.manager.yaml"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            emit_definition_deprecation_warning(def_path, "manager")

            message = str(w[0].message)
            assert "madsci migrate" in message


class TestDeprecatedDecorator:
    """Tests for the @deprecated decorator."""

    def test_emits_warning_on_call(self) -> None:
        """Test that decorated function emits warning when called."""

        @deprecated(reason="Use new_func instead")
        def old_func() -> str:
            return "result"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = old_func()

            assert result == "result"
            assert len(w) == 1
            assert issubclass(w[0].category, MadsciDeprecationWarning)
            assert "old_func" in str(w[0].message)
            assert "Use new_func instead" in str(w[0].message)

    def test_includes_replacement(self) -> None:
        """Test that warning includes replacement suggestion."""

        @deprecated(reason="Obsolete", replacement="new_function")
        def old_function() -> None:
            pass

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            old_function()

            message = str(w[0].message)
            assert "new_function" in message

    def test_includes_version_info(self) -> None:
        """Test that warning includes version information."""

        @deprecated(reason="Test", deprecated_in="0.5.0", removal_in="0.6.0")
        def versioned_func() -> None:
            pass

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            versioned_func()

            message = str(w[0].message)
            assert "0.5.0" in message
            assert "0.6.0" in message

    def test_preserves_function_metadata(self) -> None:
        """Test that decorator preserves function metadata."""

        @deprecated(reason="Test")
        def documented_func() -> None:
            """This is my docstring."""

        assert documented_func.__name__ == "documented_func"
        assert documented_func.__doc__ == "This is my docstring."

    def test_passes_arguments(self) -> None:
        """Test that decorator passes arguments correctly."""

        @deprecated(reason="Test")
        def func_with_args(a: int, b: str, c: bool = False) -> tuple:
            return (a, b, c)

        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            result = func_with_args(1, "hello", c=True)

            assert result == (1, "hello", True)


class TestDeprecatedParameterDecorator:
    """Tests for the @deprecated_parameter decorator."""

    def test_warns_when_parameter_used(self) -> None:
        """Test that warning is emitted when deprecated parameter is used."""

        @deprecated_parameter("old_param", reason="Use new_param")
        def my_func(
            new_param: Optional[str] = None, old_param: Optional[str] = None
        ) -> str:
            return old_param or new_param or "default"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = my_func(old_param="value")

            assert result == "value"
            assert len(w) == 1
            assert issubclass(w[0].category, MadsciDeprecationWarning)
            assert "old_param" in str(w[0].message)

    def test_no_warning_when_parameter_not_used(self) -> None:
        """Test that no warning is emitted when deprecated parameter is not used."""

        @deprecated_parameter("old_param", reason="Use new_param")
        def my_func(
            new_param: Optional[str] = None,
            old_param: Optional[str] = None,  # noqa: ARG001
        ) -> str:
            return new_param or "default"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = my_func(new_param="value")

            assert result == "value"
            assert len(w) == 0

    def test_no_warning_when_parameter_is_none(self) -> None:
        """Test that no warning is emitted when parameter is explicitly None."""

        @deprecated_parameter("old_param", reason="Use new_param")
        def my_func(
            new_param: Optional[str] = None,
            old_param: Optional[str] = None,  # noqa: ARG001
        ) -> str:
            return new_param or "default"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = my_func(old_param=None)

            assert result == "default"
            assert len(w) == 0

    def test_includes_replacement(self) -> None:
        """Test that warning includes replacement suggestion."""

        @deprecated_parameter("old_param", reason="Obsolete", replacement="new_param")
        def my_func(
            new_param: Optional[str] = None, old_param: Optional[str] = None
        ) -> str:
            return old_param or new_param or "default"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            my_func(old_param="value")

            message = str(w[0].message)
            assert "new_param" in message


class TestMadsciDeprecationWarning:
    """Tests for the MadsciDeprecationWarning class."""

    def test_is_deprecation_warning_subclass(self) -> None:
        """Test that MadsciDeprecationWarning is a DeprecationWarning subclass."""
        assert issubclass(MadsciDeprecationWarning, DeprecationWarning)

    def test_warning_not_filtered_by_default(self) -> None:
        """Test that MadsciDeprecationWarning is not filtered by default."""
        # This test verifies the warning filter is set up correctly
        with warnings.catch_warnings(record=True) as w:
            # Use "default" to match the filter we set
            warnings.filterwarnings("default", category=MadsciDeprecationWarning)
            warnings.warn("Test warning", MadsciDeprecationWarning, stacklevel=2)

            assert len(w) == 1
            assert issubclass(w[0].category, MadsciDeprecationWarning)



"""Pytest unit tests for the MongoDBVersionChecker with schema versioning."""

import json
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from madsci.common.mongodb_version_checker import MongoDBVersionChecker
from pydantic_extra_types.semantic_version import SemanticVersion


@pytest.fixture
def temp_schema_file(tmp_path):
    """Create a temporary schema file for testing."""
    schema_file = tmp_path / "schema.json"
    schema_content = {
        "database": "test_db",
        "schema_version": "1.0.0",
        "description": "Test schema",
        "collections": {
            "test_collection": {"description": "Test collection", "indexes": []},
            "schema_versions": {
                "description": "Version tracking",
                "indexes": [
                    {
                        "keys": [["version", 1]],
                        "name": "version_unique",
                        "unique": True,
                    }
                ],
            },
        },
    }
    schema_file.write_text(json.dumps(schema_content))
    return str(schema_file)


@pytest.fixture
def mock_mongo_client():
    """Mock MongoDB client for testing."""
    with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
        mock_db = Mock()
        mock_collection = Mock()

        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        mock_db.list_collection_names.return_value = [
            "test_collection",
            "schema_versions",
        ]
        mock_db.__getitem__ = Mock(return_value=mock_collection)

        yield mock_client, mock_db, mock_collection


class TestSchemaVersionComparison:
    """Test schema version comparison logic in is_migration_needed."""

    def test_migration_needed_for_patch_version_difference(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test that patch version differences DO trigger migration (schema versions must match exactly)."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has 1.0.0, Schema expects 1.0.1
        mock_collection.find_one.return_value = {"version": "1.0.0"}

        # Update temp schema file to have version 1.0.1
        schema_path = Path(temp_schema_file)
        schema_content = json.loads(schema_path.read_text())
        schema_content["schema_version"] = "1.0.1"
        schema_path.write_text(json.dumps(schema_content))

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("1.0.1")
        assert db_version == SemanticVersion.parse("1.0.0")

    def test_migration_needed_for_prerelease_version_difference(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test that pre-release version differences DO trigger migration (schema versions must match exactly)."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has 1.0.0, Schema expects 1.0.0-rc1
        mock_collection.find_one.return_value = {"version": "1.0.0"}

        # Update temp schema file to have version 1.0.0-rc1
        schema_path = Path(temp_schema_file)
        schema_content = json.loads(schema_path.read_text())
        schema_content["schema_version"] = "1.0.0-rc1"
        schema_path.write_text(json.dumps(schema_content))

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("1.0.0-rc1")
        assert db_version == SemanticVersion.parse("1.0.0")

    def test_migration_needed_for_minor_version_difference(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test that minor version differences trigger migration."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has 1.0.0, Schema expects 1.1.0
        mock_collection.find_one.return_value = {"version": "1.0.0"}

        # Update temp schema file to have version 1.1.0
        schema_path = Path(temp_schema_file)
        schema_content = json.loads(schema_path.read_text())
        schema_content["schema_version"] = "1.1.0"
        schema_path.write_text(json.dumps(schema_content))

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("1.1.0")
        assert db_version == SemanticVersion.parse("1.0.0")

    def test_migration_needed_for_major_version_difference(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test that major version differences trigger migration."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has 1.0.0, Schema expects 2.0.0
        mock_collection.find_one.return_value = {"version": "1.0.0"}

        # Update temp schema file to have version 2.0.0
        schema_path = Path(temp_schema_file)
        schema_content = json.loads(schema_path.read_text())
        schema_content["schema_version"] = "2.0.0"
        schema_path.write_text(json.dumps(schema_content))

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("2.0.0")
        assert db_version == SemanticVersion.parse("1.0.0")

    def test_no_migration_for_exact_version_match(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test that exact version matches don't trigger migration."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database and Schema both have 1.0.0
        mock_collection.find_one.return_value = {"version": "1.0.0"}

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        assert needs_migration is False
        assert expected_schema_version == SemanticVersion.parse("1.0.0")
        assert db_version == SemanticVersion.parse("1.0.0")

    def test_migration_needed_when_db_version_newer(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test that migration is needed when DB version is newer than expected schema version."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has 1.0.2, Schema expects 1.0.1
        mock_collection.find_one.return_value = {"version": "1.0.2"}

        # Update temp schema file to have version 1.0.1
        schema_path = Path(temp_schema_file)
        schema_content = json.loads(schema_path.read_text())
        schema_content["schema_version"] = "1.0.1"
        schema_path.write_text(json.dumps(schema_content))

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("1.0.1")
        assert db_version == SemanticVersion.parse("1.0.2")

    def test_fallback_for_invalid_db_versions(
        self, mock_mongo_client, temp_schema_file
    ):
        """Test behavior when database has invalid semantic version."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has invalid semantic version
        mock_collection.find_one.return_value = {"version": "invalid-version"}

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        # Should trigger migration due to version comparison failure (returns None on error)
        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("1.0.0")
        assert db_version is None

    def test_complex_prerelease_versions(self, mock_mongo_client, temp_schema_file):
        """Test complex pre-release version handling."""
        _mock_client, _mock_db, mock_collection = mock_mongo_client

        # Database has 1.0.0-alpha.1, Schema expects 1.0.0-beta.2
        mock_collection.find_one.return_value = {"version": "1.0.0-alpha.1"}

        # Update temp schema file to have version 1.0.0-beta.2
        schema_path = Path(temp_schema_file)
        schema_content = json.loads(schema_path.read_text())
        schema_content["schema_version"] = "1.0.0-beta.2"
        schema_path.write_text(json.dumps(schema_content))

        checker = MongoDBVersionChecker(
            "mongodb://localhost:27017", "test_db", temp_schema_file
        )

        needs_migration, expected_schema_version, db_version = (
            checker.is_migration_needed()
        )

        # Pre-release differences should trigger migration (exact match required)
        assert needs_migration is True
        assert expected_schema_version == SemanticVersion.parse("1.0.0-beta.2")
        assert db_version == SemanticVersion.parse("1.0.0-alpha.1")


class TestExistingFunctionality:
    """Test that existing functionality still works."""

    def test_migration_needed_for_no_version_tracking(self, temp_schema_file):
        """Test that databases without version tracking trigger migration."""
        with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
            mock_db = Mock()
            mock_client_instance = Mock()
            mock_client_instance.__getitem__ = Mock(return_value=mock_db)
            mock_client.return_value = mock_client_instance

            # Database exists but no schema_versions collection
            mock_db.list_collection_names.return_value = ["test_collection"]

            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "test_db", temp_schema_file
            )

            needs_migration, expected_schema_version, db_version = (
                checker.is_migration_needed()
            )

            assert needs_migration is True
            assert expected_schema_version == SemanticVersion.parse("1.0.0")
            assert db_version == SemanticVersion(0, 0, 0)

    def test_migration_needed_for_nonexistent_database(self, temp_schema_file):
        """Test that non-existent databases trigger migration."""
        with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
            mock_db = Mock()
            mock_client_instance = Mock()
            mock_client_instance.__getitem__ = Mock(return_value=mock_db)
            mock_client.return_value = mock_client_instance

            # Database doesn't exist (no collections)
            mock_db.list_collection_names.return_value = []

            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "test_db", temp_schema_file
            )

            needs_migration, expected_schema_version, db_version = (
                checker.is_migration_needed()
            )

            assert needs_migration is True
            assert expected_schema_version == SemanticVersion.parse("1.0.0")
            assert db_version is None



"""Standalone MongoDB backup and restore tool."""

import hashlib
import json
import shutil
import subprocess
import time
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from madsci.client.event_client import EventClient
from madsci.common.backup_tools.backup_manager import BackupManager
from madsci.common.backup_tools.backup_validator import BackupValidator
from madsci.common.backup_tools.base_backup import AbstractBackupTool, BackupInfo
from madsci.common.types.backup_types import MongoDBBackupSettings
from madsci.common.types.event_types import EventType
from pymongo import MongoClient


class MongoDBBackupTool(AbstractBackupTool):
    """Standalone MongoDB backup and restore tool."""

    def __init__(
        self, settings: MongoDBBackupSettings, logger: Optional[EventClient] = None
    ) -> None:
        """
        Initialize MongoDB backup tool.

        Args:
            settings: MongoDB backup configuration settings
            logger: Optional logger instance
        """
        self.settings = settings
        self.logger = logger or EventClient()
        self.backup_dir = self._setup_backup_dir()
        self.validator = BackupValidator()
        self.manager = BackupManager()

        # Initialize MongoDB connection
        self.client = MongoClient(str(settings.mongo_db_url))
        self.database = self.client[settings.database]

    def _setup_backup_dir(self) -> Path:
        """Set up backup directory, creating it if it doesn't exist."""
        backup_dir = self.settings.backup_dir
        if not backup_dir.is_absolute():
            backup_dir = Path.cwd() / backup_dir

        backup_dir.mkdir(parents=True, exist_ok=True)
        return backup_dir

    def create_backup(self, name_suffix: Optional[str] = None) -> Path:
        """
        Create a MongoDB backup using mongodump.

        Args:
            name_suffix: Optional suffix to add to backup name

        Returns:
            Path to the created backup directory

        Raises:
            RuntimeError: If backup creation fails
        """
        backup_path = self._generate_backup_path(name_suffix)
        mongodump_cmd = self._build_mongodump_command(backup_path)

        # Execute mongodump command
        try:
            self._execute_backup_command(mongodump_cmd, backup_path)
        except FileNotFoundError as fe:
            raise RuntimeError(
                "mongodump command not found. Please ensure MongoDB tools are installed."
            ) from fe
        except subprocess.CalledProcessError as e:
            self._cleanup_failed_backup(backup_path, e)
            raise RuntimeError(f"Database backup failed: {e}") from e

        # Post-process backup (metadata, validation, rotation)
        try:
            self._post_backup_processing(backup_path)
        except Exception as e:
            self._cleanup_failed_backup(backup_path, e)
            raise RuntimeError(f"Backup post-processing failed: {e}") from e

        return backup_path

    def _generate_backup_path(self, name_suffix: Optional[str]) -> Path:
        """Generate backup path with timestamp and optional suffix."""
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S_%f")
        suffix = f"_{name_suffix}" if name_suffix else ""
        backup_filename = f"{self.settings.database}_backup_{timestamp}{suffix}"
        return self.backup_dir / backup_filename

    def _build_mongodump_command(self, backup_path: Path) -> List[str]:
        """Build the mongodump command with all necessary parameters."""
        mongodump_cmd = ["mongodump"]

        # Add connection parameters
        parsed_url = self.settings.mongo_db_url
        if parsed_url.host:
            port = parsed_url.port or 27017
            mongodump_cmd.extend(["--host", f"{parsed_url.host}:{port}"])

        if parsed_url.username:
            mongodump_cmd.extend(["--username", parsed_url.username])

        if parsed_url.password:
            mongodump_cmd.extend(["--password", parsed_url.password])

        # Specify database and output directory
        mongodump_cmd.extend(
            ["--db", self.settings.database, "--out", str(backup_path)]
        )

        # Add collection filters if specified
        if self.settings.collections:
            for collection in self.settings.collections:
                mongodump_cmd.extend(["--collection", collection])

        return mongodump_cmd

    def _execute_backup_command(
        self, mongodump_cmd: List[str], backup_path: Path
    ) -> None:
        """Execute the mongodump command and validate success."""
        # Ensure backup directory exists (mongodump requires parent directory to exist)
        backup_path.mkdir(parents=True, exist_ok=True)

        self.logger.info(
            "Creating database backup",
            event_type=EventType.BACKUP_CREATE,
            backup_path=str(backup_path),
            database=self.settings.database,
        )
        result = subprocess.run(  # noqa: S603
            mongodump_cmd, capture_output=True, text=True, check=True
        )

        if result.returncode == 0:
            self.logger.info(
                "Database backup completed successfully",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
                database=self.settings.database,
            )

            # Verify backup directory was created
            if not backup_path.exists():
                raise RuntimeError(
                    f"mongodump reported success but backup directory was not created: {backup_path}. "
                    f"This may indicate a connection issue or empty database."
                )

            # Verify database subdirectory exists
            db_backup_path = backup_path / self.settings.database
            if not db_backup_path.exists():
                # Check if database is empty
                try:
                    collections = list(self.database.list_collection_names())
                    if not collections:
                        raise RuntimeError(
                            f"Cannot backup database '{self.settings.database}': database is empty (no collections). "
                            f"MongoDB does not create backup files for empty databases."
                        )
                except RuntimeError:
                    # Re-raise our own RuntimeError about empty database
                    raise
                except Exception as e:
                    # Log connection errors but continue with generic message
                    self.logger.warning(
                        "Could not check if database is empty",
                        event_type=EventType.BACKUP_CREATE,
                        database=self.settings.database,
                        error=str(e),
                        exc_info=True,
                    )

                raise RuntimeError(
                    f"mongodump created backup directory but database subdirectory is missing: {db_backup_path}. "
                    f"Check that database '{self.settings.database}' exists and is accessible."
                )
        else:
            raise RuntimeError(f"mongodump failed: {result.stderr}")

    def _post_backup_processing(self, backup_path: Path) -> None:
        """Handle post-backup validation, metadata creation, and rotation."""
        # Create metadata first (required for validation)
        self._create_backup_metadata(backup_path)

        # Validate backup integrity if enabled
        if self.settings.validate_integrity:
            self._validate_backup_integrity(backup_path)

        # Rotate old backups
        if self.settings.max_backups > 0:
            self.manager.rotate_backups(self.backup_dir, self.settings.max_backups)

    def _cleanup_failed_backup(self, backup_path: Path, error: Exception) -> None:
        """Clean up after a failed backup operation."""
        self.logger.error(
            "Backup failed",
            event_type=EventType.BACKUP_CREATE,
            backup_path=str(backup_path),
            database=self.settings.database,
            error=str(error),
        )
        if backup_path.exists():
            shutil.rmtree(backup_path)

    def restore_from_backup(
        self, backup_path: Path, target_db: Optional[str] = None
    ) -> None:
        """
        Restore database from a backup directory using mongorestore.

        Args:
            backup_path: Path to backup directory
            target_db: Optional target database name (defaults to original database)

        Raises:
            FileNotFoundError: If backup directory doesn't exist
            RuntimeError: If restore operation fails
        """
        if not backup_path.exists():
            raise FileNotFoundError(f"Backup directory not found: {backup_path}")

        # The backup path should contain a subdirectory with the database name
        db_backup_path = backup_path / self.settings.database
        if not db_backup_path.exists():
            raise FileNotFoundError(f"Database backup not found in: {db_backup_path}")

        # Use target database if specified, otherwise use original database
        restore_db = target_db or self.settings.database

        # Build mongorestore command
        mongorestore_cmd = ["mongorestore"]

        # Add connection parameters
        parsed_url = self.settings.mongo_db_url
        if parsed_url.host:
            port = parsed_url.port or 27017
            mongorestore_cmd.extend(["--host", f"{parsed_url.host}:{port}"])

        if parsed_url.username:
            mongorestore_cmd.extend(["--username", parsed_url.username])

        if parsed_url.password:
            mongorestore_cmd.extend(["--password", parsed_url.password])

        # Drop existing database and restore
        mongorestore_cmd.extend(
            [
                "--drop",  # Drop existing collections before restoring
                "--db",
                restore_db,
                str(db_backup_path),
            ]
        )

        try:
            self.logger.info(
                "Restoring database from backup",
                event_type=EventType.BACKUP_RESTORE,
                backup_path=str(backup_path),
                database=restore_db,
                source_database=self.settings.database,
            )
            result = subprocess.run(  # noqa: S603
                mongorestore_cmd, capture_output=True, text=True, check=True
            )

            if result.returncode == 0:
                self.logger.info(
                    "Database restore completed successfully",
                    event_type=EventType.BACKUP_RESTORE,
                    database=restore_db,
                    source_database=self.settings.database,
                )
                # Verify restore success
                if not self._verify_restore_success(backup_path):
                    raise RuntimeError("Restore verification failed")
            else:
                raise RuntimeError(f"mongorestore failed: {result.stderr}")

        except subprocess.CalledProcessError as e:
            self.logger.error(
                "Restore failed",
                event_type=EventType.BACKUP_RESTORE,
                backup_path=str(backup_path),
                database=restore_db,
                source_database=self.settings.database,
                stderr=str(e.stderr),
                exc_info=True,
            )
            self._cleanup_failed_restore(backup_path)
            raise RuntimeError(f"Database restore failed: {e}") from e

    def validate_backup_integrity(self, backup_path: Path) -> bool:
        """
        Validate backup integrity using checksums and restore testing.

        Args:
            backup_path: Path to backup directory

        Returns:
            True if backup is valid, False otherwise
        """
        try:
            return self._validate_backup_integrity(backup_path)
        except Exception as e:
            self.logger.error(
                "Backup validation failed",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
                database=self.settings.database,
                error=str(e),
                exc_info=True,
            )
            return False

    def list_available_backups(self) -> List[BackupInfo]:
        """
        List available backups with metadata.

        Returns:
            List of BackupInfo objects for available backups
        """
        backups = []

        if not self.backup_dir.exists():
            return backups

        for item in self.backup_dir.iterdir():
            if item.is_dir() and (item / self.settings.database).exists():
                # Try to load metadata
                metadata_file = item / "backup_metadata.json"
                if metadata_file.exists():
                    try:
                        metadata = json.loads(metadata_file.read_text())

                        backup_info = BackupInfo(
                            backup_path=item,
                            created_at=datetime.fromisoformat(metadata["timestamp"]),
                            database_version=metadata.get("database_version"),
                            backup_size=metadata["backup_size"],
                            checksum=metadata["checksum"],
                            backup_type="mongodb",
                            is_valid=self._check_backup_validity(item),
                        )
                        backups.append(backup_info)
                    except (json.JSONDecodeError, KeyError) as e:
                        self.logger.warning(
                            "Invalid metadata for backup",
                            event_type=EventType.BACKUP_CREATE,
                            backup_path=str(item),
                            error=str(e),
                            exc_info=True,
                        )

        # Sort by creation time (newest first)
        backups.sort(key=lambda x: x.created_at, reverse=True)
        return backups

    def delete_backup(self, backup_path: Path) -> None:
        """
        Delete a specific backup directory.

        Args:
            backup_path: Path to backup directory to delete

        Raises:
            RuntimeError: If deletion fails
        """
        try:
            if backup_path.exists():
                shutil.rmtree(backup_path)
                self.logger.info(
                    "Deleted backup",
                    event_type=EventType.BACKUP_CREATE,
                    backup_path=str(backup_path),
                )
            else:
                self.logger.warning(
                    "Backup path does not exist",
                    event_type=EventType.BACKUP_CREATE,
                    backup_path=str(backup_path),
                )
        except Exception as e:
            raise RuntimeError(f"Failed to delete backup {backup_path}: {e}") from e

    def _validate_backup_integrity(self, backup_path: Path) -> bool:
        """Perform comprehensive MongoDB backup validation."""
        self.logger.info(
            "Validating backup integrity",
            event_type=EventType.BACKUP_CREATE,
            backup_path=str(backup_path),
            database=self.settings.database,
        )

        # Step 1: Verify backup completion
        if not self._verify_backup_completion(backup_path):
            raise RuntimeError(f"Backup completion verification failed: {backup_path}")

        # Step 2: Check existing checksum (if this is validation of existing backup)
        checksum_file = backup_path / "backup.checksum"
        if checksum_file.exists():
            # Validate existing checksum
            if not self._validate_backup_checksum(backup_path):
                raise RuntimeError(f"Backup checksum validation failed: {backup_path}")
        else:
            # Generate new checksum (if this is backup creation)
            self._generate_backup_checksum(backup_path)

        # Step 3: Test restorability
        if not self._test_backup_restore(backup_path):
            raise RuntimeError(f"Backup restore test failed: {backup_path}")

        self.logger.info(
            "Backup integrity validation passed",
            event_type=EventType.BACKUP_CREATE,
            backup_path=str(backup_path),
            database=self.settings.database,
        )
        return True

    def _verify_backup_completion(self, backup_path: Path) -> bool:
        """Verify that MongoDB backup completed successfully."""
        if not backup_path.exists():
            return False

        # Check if backup directory contains database subdirectory
        db_backup_path = backup_path / self.settings.database
        if not db_backup_path.exists():
            return False

        # Check if backup contains collection files
        bson_files = list(db_backup_path.glob("*.bson"))
        if not bson_files:
            self.logger.warning(
                "No BSON files found in backup directory",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
                database=self.settings.database,
            )
            return False

        # Check if backup contains metadata
        metadata_files = list(db_backup_path.glob("*.metadata.json"))
        if not metadata_files:
            self.logger.warning(
                "No metadata files found in backup directory",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
                database=self.settings.database,
            )

        self.logger.info(
            "Backup verification successful",
            event_type=EventType.BACKUP_CREATE,
            backup_path=str(backup_path),
            database=self.settings.database,
            collections_count=len(bson_files),
        )
        return True

    def _generate_backup_checksum(self, backup_path: Path) -> str:
        """Generate SHA256 checksum for MongoDB backup directory."""
        checksum = self._generate_backup_checksum_inline(backup_path)

        # Save checksum to file
        checksum_file = backup_path / "backup.checksum"
        checksum_file.write_text(checksum)

        self.logger.info(
            "Generated backup checksum",
            event_type=EventType.BACKUP_CREATE,
            backup_path=str(backup_path),
            checksum=checksum,
        )
        return checksum

    def _generate_backup_checksum_inline(self, backup_path: Path) -> str:
        """Generate checksum without saving to file (for validation)."""
        sha256_hash = hashlib.sha256()

        # Generate checksum based on all BSON files in the backup
        db_backup_path = backup_path / self.settings.database
        bson_files = sorted(db_backup_path.glob("*.bson"))

        for bson_file in bson_files:
            with bson_file.open("rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    sha256_hash.update(chunk)

        return sha256_hash.hexdigest()

    def _validate_backup_checksum(self, backup_path: Path) -> bool:
        """Validate MongoDB backup directory against its checksum."""
        checksum_file = backup_path / "backup.checksum"

        if not checksum_file.exists():
            self.logger.warning(
                "No checksum file found for backup validation",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
            )
            return False

        try:
            expected_checksum = checksum_file.read_text().strip()
            actual_checksum = self._generate_backup_checksum_inline(backup_path)

            if expected_checksum == actual_checksum:
                self.logger.info(
                    "Backup checksum validation passed",
                    event_type=EventType.BACKUP_CREATE,
                    backup_path=str(backup_path),
                )
                return True

            self.logger.error(
                "Backup checksum validation failed",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
                expected_checksum=expected_checksum,
                actual_checksum=actual_checksum,
            )
            return False

        except Exception as e:
            self.logger.error(
                "Error validating backup checksum",
                event_type=EventType.BACKUP_CREATE,
                backup_path=str(backup_path),
                error=str(e),
                exc_info=True,
            )
            return False

    def _test_backup_restore(self, backup_path: Path) -> bool:
        """Test MongoDB backup by attempting restore to temporary database."""
        test_db_name = f"test_restore_{int(time.time())}"

        try:
            # Build mongorestore command for test database
            mongorestore_cmd = ["mongorestore"]

            # Add connection parameters
            parsed_url = self.settings.mongo_db_url
            if parsed_url.host:
                port = parsed_url.port or 27017
                mongorestore_cmd.extend(["--host", f"{parsed_url.host}:{port}"])

            if parsed_url.username:
                mongorestore_cmd.extend(["--username", parsed_url.username])

            if parsed_url.password:
                mongorestore_cmd.extend(["--password", parsed_url.password])

            # Restore to test database
            db_backup_path = backup_path / self.settings.database
            mongorestore_cmd.extend(["--db", test_db_name, str(db_backup_path)])

            result = subprocess.run(  # noqa: S603
                mongorestore_cmd,
                check=False,
                capture_output=True,
                text=True,
                timeout=300,
            )

            success = result.returncode == 0
            if success:
                self.logger.info(
                    "Backup restore test successful",
                    event_type=EventType.BACKUP_RESTORE,
                    backup_path=str(backup_path),
                    test_db=test_db_name,
                )
            else:
                self.logger.error(
                    "Backup restore test failed",
                    event_type=EventType.BACKUP_RESTORE,
                    backup_path=str(backup_path),
                    test_db=test_db_name,
                    stderr=str(result.stderr),
                )

            return success

        except Exception as e:
            self.logger.error(
                "Error testing backup restore",
                event_type=EventType.BACKUP_RESTORE,
                backup_path=str(backup_path),
                test_db=test_db_name,
                error=str(e),
                exc_info=True,
            )
            return False
        finally:
            # Clean up test database
            try:
                test_client = MongoClient(str(self.settings.mongo_db_url))
                test_client.drop_database(test_db_name)
                test_client.close()
            except Exception as e:
                self.logger.warning(
                    "Failed to clean up test database",
                    event_type=EventType.BACKUP_RESTORE,
                    test_db=test_db_name,
                    error=str(e),
                    exc_info=True,
                )

    def _create_backup_metadata(self, backup_path: Path) -> None:
        """Create metadata file for MongoDB backup."""
        checksum = self._generate_backup_checksum_inline(backup_path)

        metadata = {
            "timestamp": datetime.now().isoformat(),
            "database_version": None,  # MongoDB doesn't have schema versions like PostgreSQL
            "backup_size": sum(
                f.stat().st_size for f in backup_path.rglob("*") if f.is_file()
            ),
            "checksum": checksum,
            "database_name": self.settings.database,
            "collections_count": len(
                list((backup_path / self.settings.database).glob("*.bson"))
            ),
        }

        metadata_file = backup_path / "backup_metadata.json"
        metadata_file.write_text(json.dumps(metadata, indent=2))

        self.logger.info(
            "Created backup metadata",
            event_type=EventType.BACKUP_CREATE,
            metadata_file=str(metadata_file),
            backup_path=str(backup_path),
        )

        # Save checksum to separate file (required for validation)
        checksum_file = backup_path / "backup.checksum"
        checksum_file.write_text(checksum)
        self.logger.info(
            "Created backup checksum",
            event_type=EventType.BACKUP_CREATE,
            checksum_file=str(checksum_file),
            backup_path=str(backup_path),
        )

    def _verify_restore_success(self, backup_path: Path) -> bool:
        """Verify that MongoDB restore operation was successful."""
        try:
            # Check that collections exist and have expected structure
            db_backup_path = backup_path / self.settings.database
            bson_files = list(db_backup_path.glob("*.bson"))

            for bson_file in bson_files:
                collection_name = bson_file.stem
                # Verify collection exists in database
                if collection_name not in self.database.list_collection_names():
                    self.logger.error(
                        "Collection not found after restore",
                        event_type=EventType.BACKUP_RESTORE,
                        collection=collection_name,
                    )
                    return False

                # Basic document count check
                doc_count = self.database[collection_name].count_documents({})
                self.logger.info(
                    "Collection has documents after restore",
                    event_type=EventType.BACKUP_RESTORE,
                    collection=collection_name,
                    doc_count=doc_count,
                )

            self.logger.info(
                "Restore verification successful",
                event_type=EventType.BACKUP_RESTORE,
                database=self.settings.database,
            )
            return True

        except Exception as e:
            self.logger.error(
                "Error verifying restore success",
                event_type=EventType.BACKUP_RESTORE,
                error=str(e),
                exc_info=True,
            )
            return False

    def _cleanup_failed_restore(self, backup_path: Path) -> None:  # noqa: ARG002
        """Cleanup after a failed MongoDB restore operation."""
        try:
            # Drop all collections to clean state
            for collection_name in self.database.list_collection_names():
                self.database.drop_collection(collection_name)
                self.logger.info(
                    "Dropped collection during cleanup",
                    event_type=EventType.BACKUP_RESTORE,
                    collection=collection_name,
                )

            self.logger.info(
                "Failed restore cleanup completed",
                event_type=EventType.BACKUP_RESTORE,
                database=self.settings.database,
            )

        except Exception as e:
            self.logger.error(
                "Error during failed restore cleanup",
                event_type=EventType.BACKUP_RESTORE,
                error=str(e),
                exc_info=True,
            )

    def _check_backup_validity(self, backup_path: Path) -> bool:
        """Check if a backup is valid by verifying its checksum."""
        try:
            return self._validate_backup_checksum(backup_path)
        except Exception:
            return False



"""Tests for the .madsci/ sentry directory resolution module.

Tests cover:
- find_madsci_dir walk-up, boundaries, fallback
- get_madsci_subdir creation and delegation
- ensure_madsci_dir scaffolding
- get_global_madsci_subdir home-only behaviour
- _resolve_sentry_paths model validator on MadsciBaseSettings
"""

from __future__ import annotations

import json
from pathlib import Path

import pytest
from madsci.common.sentry import (
    REGISTRY_FILE,
    SENTRY_DIR_NAME,
    STANDARD_SUBDIRS,
    ensure_madsci_dir,
    find_madsci_dir,
    get_global_madsci_subdir,
    get_madsci_subdir,
)
from madsci.common.types.base_types import MadsciBaseSettings
from pydantic_settings import SettingsConfigDict

# ---------------------------------------------------------------------------
# find_madsci_dir
# ---------------------------------------------------------------------------


class TestFindMadsciDir:
    """Tests for the canonical .madsci/ resolution function."""

    def test_finds_madsci_in_start_dir(self, tmp_path: Path) -> None:
        """Direct .madsci/ in start_dir is found immediately."""
        (tmp_path / SENTRY_DIR_NAME).mkdir()
        result = find_madsci_dir(tmp_path)
        assert result == tmp_path / SENTRY_DIR_NAME

    def test_walks_up_to_find_madsci(self, tmp_path: Path) -> None:
        """Walk up from child to find .madsci/ in parent."""
        child = tmp_path / "a" / "b"
        child.mkdir(parents=True)
        (tmp_path / SENTRY_DIR_NAME).mkdir()

        result = find_madsci_dir(child)
        assert result == tmp_path / SENTRY_DIR_NAME

    def test_stops_at_madsci_sentinel(self, tmp_path: Path) -> None:
        """Walk-up stops at the first .madsci/ found."""
        project = tmp_path / "project"
        child = project / "sub"
        child.mkdir(parents=True)
        (project / SENTRY_DIR_NAME).mkdir()
        # Another .madsci/ higher up should not be found
        (tmp_path / SENTRY_DIR_NAME).mkdir()

        result = find_madsci_dir(child)
        assert result == project / SENTRY_DIR_NAME

    def test_git_boundary_fallback(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """When no .madsci/ exists, .git/ serves as a boundary."""
        fakehome = tmp_path / "fakehome"
        project = fakehome / "project"
        child = project / "sub"
        child.mkdir(parents=True)
        (project / ".git").mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(child)
        assert result == project / SENTRY_DIR_NAME

    def test_madsci_takes_priority_over_git(self, tmp_path: Path) -> None:
        """.madsci/ takes priority over .git/ when both exist."""
        project = tmp_path / "project"
        child = project / "sub"
        child.mkdir(parents=True)
        (project / SENTRY_DIR_NAME).mkdir()
        (project / ".git").mkdir()

        result = find_madsci_dir(child)
        assert result == project / SENTRY_DIR_NAME

    def test_falls_back_to_home(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Falls back to ~/.madsci/ when neither .madsci/ nor .git/ found."""
        fakehome = tmp_path / "fakehome"
        child = fakehome / "project" / "sub"
        child.mkdir(parents=True)
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(child)
        assert result == fakehome / SENTRY_DIR_NAME

    def test_respects_max_levels(self, tmp_path: Path) -> None:
        """max_levels limits how far up the walk goes."""
        deep = tmp_path / "a" / "b" / "c" / "d"
        deep.mkdir(parents=True)
        (tmp_path / SENTRY_DIR_NAME).mkdir()

        # max_levels=2 should not reach tmp_path (4 levels up)
        result = find_madsci_dir(deep, max_levels=2)
        # Should fall back to home
        assert result == Path.home() / SENTRY_DIR_NAME

    def test_respects_home_boundary(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Walk-up stops at home directory."""
        fakehome = tmp_path / "fakehome"
        child = fakehome / "project"
        child.mkdir(parents=True)
        # .madsci/ above home — should not be found via walk-up
        (tmp_path / SENTRY_DIR_NAME).mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(child)
        # Falls back to ~/.madsci/ (not the one above home)
        assert result == fakehome / SENTRY_DIR_NAME

    def test_respects_settings_dir_env_var(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """MADSCI_SETTINGS_DIR env var overrides CWD as start."""
        project = tmp_path / "my_project"
        project.mkdir()
        (project / SENTRY_DIR_NAME).mkdir()
        monkeypatch.setenv("MADSCI_SETTINGS_DIR", str(project))

        result = find_madsci_dir()
        assert result == project / SENTRY_DIR_NAME

    def test_auto_create_creates_dir(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """auto_create=True creates the .madsci/ dir when not found."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(fakehome, auto_create=True)
        assert result == fakehome / SENTRY_DIR_NAME
        assert result.is_dir()

    def test_auto_create_false_does_not_create(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """auto_create=False does not create directories."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(fakehome, auto_create=False)
        assert result == fakehome / SENTRY_DIR_NAME
        assert not result.is_dir()

    def test_git_boundary_auto_create(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """auto_create creates .madsci/ when inferred from .git/ boundary."""
        fakehome = tmp_path / "fakehome"
        project = fakehome / "project"
        project.mkdir(parents=True)
        (project / ".git").mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(project, auto_create=True)
        assert result == project / SENTRY_DIR_NAME
        assert result.is_dir()

    def test_git_file_boundary(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """A .git *file* (worktree) is treated the same as a .git/ directory."""
        fakehome = tmp_path / "fakehome"
        project = fakehome / "project"
        child = project / "sub"
        child.mkdir(parents=True)
        # .git as a file (git worktree format)
        (project / ".git").write_text("gitdir: /some/path\n")
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = find_madsci_dir(child)
        assert result == project / SENTRY_DIR_NAME


# ---------------------------------------------------------------------------
# get_madsci_subdir
# ---------------------------------------------------------------------------


class TestGetMadsciSubdir:
    """Tests for subdirectory resolution."""

    def test_returns_correct_path(self, tmp_path: Path) -> None:
        (tmp_path / SENTRY_DIR_NAME).mkdir()
        result = get_madsci_subdir("pids", tmp_path)
        assert result == tmp_path / SENTRY_DIR_NAME / "pids"

    def test_creates_when_create_true(self, tmp_path: Path) -> None:
        (tmp_path / SENTRY_DIR_NAME).mkdir()
        result = get_madsci_subdir("pids", tmp_path, create=True)
        assert result.is_dir()

    def test_does_not_create_when_create_false(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = get_madsci_subdir("pids", tmp_path, create=False)
        assert not result.is_dir()


# ---------------------------------------------------------------------------
# ensure_madsci_dir
# ---------------------------------------------------------------------------


class TestEnsureMadsciDir:
    """Tests for .madsci/ scaffolding."""

    def test_creates_standard_subdirs(self, tmp_path: Path) -> None:
        result = ensure_madsci_dir(tmp_path)
        assert result == tmp_path / SENTRY_DIR_NAME
        assert result.is_dir()
        for subdir in STANDARD_SUBDIRS:
            assert (result / subdir).is_dir()

    def test_creates_registry_json(self, tmp_path: Path) -> None:
        result = ensure_madsci_dir(tmp_path)
        registry_path = result / REGISTRY_FILE
        assert registry_path.exists()
        data = json.loads(registry_path.read_text())
        assert "entries" in data

    def test_idempotent(self, tmp_path: Path) -> None:
        """Calling twice should be safe."""
        result1 = ensure_madsci_dir(tmp_path)
        # Write something to registry to verify it's not overwritten
        registry_path = result1 / REGISTRY_FILE
        custom = {"entries": {"test": "value"}}
        registry_path.write_text(json.dumps(custom))

        result2 = ensure_madsci_dir(tmp_path)
        assert result2 == result1
        # Registry should NOT be overwritten since it already existed
        data = json.loads(registry_path.read_text())
        assert data == custom


# ---------------------------------------------------------------------------
# get_global_madsci_subdir
# ---------------------------------------------------------------------------


class TestGetGlobalMadsciSubdir:
    """Tests for the global (home-only) variant."""

    def test_uses_home_regardless_of_sentinel(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Should always use ~/.madsci/, ignoring any project sentinels."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        project = tmp_path / "project"
        project.mkdir()
        (project / SENTRY_DIR_NAME).mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = get_global_madsci_subdir("templates", create=True)
        assert result == fakehome / SENTRY_DIR_NAME / "templates"
        assert result.is_dir()

    def test_does_not_create_when_false(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        result = get_global_madsci_subdir("templates", create=False)
        assert result == fakehome / SENTRY_DIR_NAME / "templates"
        assert not result.is_dir()


# ---------------------------------------------------------------------------
# _resolve_sentry_paths model validator
# ---------------------------------------------------------------------------


class TestResolveSentryPaths:
    """Tests for the _resolve_sentry_paths model validator on MadsciBaseSettings."""

    def test_path_field_resolved_to_absolute(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """A Path field with a .madsci/ default is resolved to an absolute path."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        (tmp_path / SENTRY_DIR_NAME).mkdir()
        monkeypatch.chdir(tmp_path)
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        class MySettings(MadsciBaseSettings):
            model_config = SettingsConfigDict(env_prefix="TESTSENTRY_")
            log_dir: Path = Path(".madsci/logs")

        settings = MySettings()
        assert settings.log_dir == tmp_path / ".madsci" / "logs"
        assert settings.log_dir.is_absolute()

    def test_str_field_preserves_type(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """A str field with a .madsci/ default is resolved but stays str."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        (tmp_path / SENTRY_DIR_NAME).mkdir()
        monkeypatch.chdir(tmp_path)
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        class MySettings(MadsciBaseSettings):
            model_config = SettingsConfigDict(env_prefix="TESTSENTRY2_")
            backup_dir: str = ".madsci/backups"

        settings = MySettings()
        assert isinstance(settings.backup_dir, str)
        assert settings.backup_dir == str(tmp_path / ".madsci" / "backups")

    def test_non_sentry_path_untouched(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """A field with a non-.madsci/ path remains unchanged."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        monkeypatch.chdir(tmp_path)
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        class MySettings(MadsciBaseSettings):
            model_config = SettingsConfigDict(env_prefix="TESTSENTRY3_")
            data_dir: Path = Path("/usr/local/data")

        settings = MySettings()
        assert settings.data_dir == Path("/usr/local/data")

    def test_git_boundary_resolution(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """When no .madsci/ exists but .git/ does, resolve via git boundary."""
        fakehome = tmp_path / "fakehome"
        fakehome.mkdir()
        project = tmp_path / "project"
        project.mkdir()
        (project / ".git").mkdir()
        monkeypatch.chdir(project)
        monkeypatch.setattr(Path, "home", classmethod(lambda _cls: fakehome))

        class MySettings(MadsciBaseSettings):
            model_config = SettingsConfigDict(env_prefix="TESTSENTRY4_")
            log_dir: Path = Path(".madsci/logs")

        settings = MySettings()
        assert settings.log_dir == project / ".madsci" / "logs"
        assert settings.log_dir.is_absolute()



"""Custom serializers for use in MADSci dataclasses."""

from typing import TYPE_CHECKING, Any, Union

import yaml

if TYPE_CHECKING:
    from madsci.common.types.base_types import MadsciBaseModel


def serialize_to_yaml(model: "MadsciBaseModel") -> str:
    """Serialize a MADSci model to YAML string.

    Args:
        model: The MADSci model to serialize

    Returns:
        YAML string representation of the model

    Example:
        from madsci.common.serializers import serialize_to_yaml

        yaml_content = serialize_to_yaml(my_pydantic_model)
    """
    return yaml.dump(
        model.model_dump(mode="json"),
        indent=2,
        sort_keys=False,
    )


def dict_to_list(dct: Union[list[Any], dict[str, Any]]) -> list[Any]:
    """Converts a dictionary to a list of values.

    Example Usage:
        from pydantic import field_serializer

        serialize_nodes_to_list = field_serializer("nodes")(dict_to_list)
    """
    if isinstance(dct, list):
        return dct
    return list(dct.values())



"""Tests for Phase G.5: NodeInfo.from_config() factory method."""

from madsci.common.types.node_types import (
    NodeConfig,
    NodeDefinition,
    NodeInfo,
    NodeType,
    RestNodeConfig,
)
from madsci.common.utils import new_ulid_str


class TestNodeInfoFromConfig:
    """Tests for the new from_config() factory."""

    def test_from_config_basic(self) -> None:
        """from_config creates NodeInfo from config settings."""
        config = RestNodeConfig(
            node_name="test_node",
            module_name="test_module",
            module_version="1.0.0",
        )
        info = NodeInfo.from_config(config)
        assert info.node_name == "test_node"
        assert info.module_name == "test_module"
        assert str(info.module_version) == "1.0.0"
        assert info.config_schema is not None

    def test_from_config_with_node_type(self) -> None:
        """from_config respects node_type setting."""
        config = RestNodeConfig(
            node_name="compute_node",
            module_name="compute_module",
            module_version="2.0.0",
            node_type=NodeType.COMPUTE,
        )
        info = NodeInfo.from_config(config)
        assert info.node_type == NodeType.COMPUTE

    def test_from_config_with_node_id(self) -> None:
        """from_config respects explicit node_id."""
        custom_id = new_ulid_str()
        config = RestNodeConfig(
            node_name="my_node",
            module_name="my_module",
            module_version="1.0.0",
            node_id=custom_id,
        )
        info = NodeInfo.from_config(config)
        assert info.node_id == custom_id

    def test_from_config_defaults(self) -> None:
        """from_config provides defaults for required fields."""
        config = NodeConfig()
        info = NodeInfo.from_config(config)
        assert info.node_name == "unnamed_node"
        assert info.module_name == "unknown_module"
        assert info.node_id is not None

    def test_from_config_with_definition_fallback(self) -> None:
        """from_config uses definition values as fallback."""
        config = NodeConfig()
        node_def = NodeDefinition(
            node_name="def_node",
            module_name="def_module",
        )
        info = NodeInfo.from_config(config, node_definition=node_def)
        assert info.node_name == "def_node"  # falls back to definition
        assert info.module_name == "def_module"

    def test_from_config_overrides_definition(self) -> None:
        """from_config config fields override definition values."""
        config = RestNodeConfig(
            node_name="config_node",
            module_name="config_module",
            module_version="2.0.0",
        )
        node_def = NodeDefinition(
            node_name="def_node",
            module_name="def_module",
        )
        info = NodeInfo.from_config(config, node_definition=node_def)
        assert info.node_name == "config_node"  # config overrides def
        assert info.module_name == "config_module"

    def test_from_config_kwargs_override_all(self) -> None:
        """from_config explicit kwargs override both config and definition."""
        config = RestNodeConfig(
            node_name="config_node",
            module_name="config_module",
            module_version="1.0.0",
        )
        info = NodeInfo.from_config(
            config, node_name="explicit_name", module_name="explicit_module"
        )
        assert info.node_name == "explicit_name"
        assert info.module_name == "explicit_module"


class TestNodeInfoFromNodeDefAndConfig:
    """Tests that the legacy factory still works."""

    def test_from_node_def_and_config(self) -> None:
        """Legacy factory creates NodeInfo from definition."""
        node_def = NodeDefinition(
            node_name="legacy_node",
            module_name="legacy_module",
        )
        config = RestNodeConfig()
        info = NodeInfo.from_node_def_and_config(node_def, config)
        assert info.node_name == "legacy_node"
        assert info.config_schema is not None

    def test_from_node_def_and_config_no_config(self) -> None:
        """Legacy factory works without config."""
        node_def = NodeDefinition(
            node_name="bare_node",
            module_name="bare_module",
        )
        info = NodeInfo.from_node_def_and_config(node_def)
        assert info.node_name == "bare_node"


class TestNodeConfigIdentityFields:
    """Tests for the new identity fields on NodeConfig."""

    def test_identity_fields_optional(self) -> None:
        """Identity fields default to None."""
        config = NodeConfig()
        assert config.node_name is None
        assert config.node_id is None
        assert config.node_type is None
        assert config.module_name is None
        assert config.module_version is None

    def test_identity_fields_settable(self) -> None:
        """Identity fields can be set."""
        config = NodeConfig(
            node_name="my_node",
            node_id="01ABC",
            node_type=NodeType.DEVICE,
            module_name="my_mod",
            module_version="1.2.3",
        )
        assert config.node_name == "my_node"
        assert config.node_id == "01ABC"
        assert config.node_type == NodeType.DEVICE
        assert config.module_name == "my_mod"
        assert config.module_version == "1.2.3"



"""In-memory drop-in replacements for external services (Redis, MongoDB).

These backends enable running the full MADSci stack without Docker by providing
thread-safe, in-memory implementations that match the exact interfaces used by
managers. Data is ephemeral (in-memory), making this ideal for development,
testing, and demos.
"""



"""Comprehensive tests for PostgreSQL backup tool functionality."""

import os
import tempfile
from datetime import datetime, timezone
from pathlib import Path
from typing import Generator
from unittest.mock import MagicMock, Mock, patch

import pytest
from madsci.client.event_client import EventClient
from madsci.common.backup_tools.postgres_backup import PostgreSQLBackupTool
from madsci.common.types.backup_types import PostgreSQLBackupSettings


class TestPostgreSQLBackupTool:
    """Test PostgreSQL backup tool functionality."""

    @pytest.fixture
    def temp_dir(self) -> Generator[Path, None, None]:
        """Create temporary directory for testing."""
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)

    @pytest.fixture
    def test_db_url(self) -> str:
        """Test database URL."""
        return "postgresql://testuser:testpass@localhost:5432/testdb"

    @pytest.fixture
    def backup_settings(
        self, temp_dir: Path, test_db_url: str
    ) -> PostgreSQLBackupSettings:
        """Create PostgreSQL backup settings for testing."""
        return PostgreSQLBackupSettings(
            db_url=test_db_url,
            backup_dir=temp_dir / "backups",
            backup_format="custom",
            max_backups=5,
            validate_integrity=False,  # Disable for subprocess mock tests
            compression=True,
        )

    @pytest.fixture
    def mock_event_client(self) -> MagicMock:
        """Mock event client for testing."""
        return MagicMock(spec=EventClient)

    @pytest.fixture
    def backup_tool(
        self, backup_settings: PostgreSQLBackupSettings, mock_event_client: MagicMock
    ) -> PostgreSQLBackupTool:
        """Create PostgreSQL backup tool for testing."""
        return PostgreSQLBackupTool(backup_settings, mock_event_client)

    def test_initialization_success(self, backup_tool: PostgreSQLBackupTool):
        """Test successful backup tool initialization."""
        assert backup_tool.settings.backup_format == "custom"
        assert backup_tool.backup_dir.exists()
        assert backup_tool.logger is not None

    def test_initialization_creates_backup_directory(
        self, backup_settings: PostgreSQLBackupSettings
    ):
        """Test backup tool creates backup directory during initialization."""
        # Ensure directory doesn't exist initially
        backup_settings.backup_dir.rmdir() if backup_settings.backup_dir.exists() else None

        tool = PostgreSQLBackupTool(backup_settings)
        assert tool.backup_dir.exists()
        assert tool.backup_dir.is_dir()

    @patch("subprocess.run")
    def test_create_backup_success(
        self, mock_run: Mock, backup_tool: PostgreSQLBackupTool
    ):
        """Test successful backup creation."""
        # Setup mock subprocess response
        mock_result = Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        mock_run.return_value = mock_result

        # Create a mock backup file that will be created by pg_dump
        def create_backup_file(*args, **_kwargs):
            # Extract file path from command arguments
            cmd = args[0]
            file_arg_index = cmd.index("--file") + 1
            backup_path = Path(cmd[file_arg_index])
            backup_path.parent.mkdir(parents=True, exist_ok=True)
            backup_path.write_text("mock backup content")
            return mock_result

        mock_run.side_effect = create_backup_file

        # Test backup creation
        backup_path = backup_tool.create_backup()

        assert backup_path.exists()
        assert backup_path.name.startswith("postgres_backup_")
        assert backup_path.suffix in [".dump", ".sql"]
        # Should be called at least once for pg_dump (may be called twice for version check)
        assert mock_run.call_count >= 1

    @patch("subprocess.run")
    def test_create_backup_with_custom_name_suffix(
        self, mock_run: Mock, backup_tool: PostgreSQLBackupTool
    ):
        """Test backup creation with custom name suffix."""
        # Setup mock subprocess response
        mock_result = Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""

        def create_backup_file(*args, **_kwargs):
            cmd = args[0]
            file_arg_index = cmd.index("--file") + 1
            backup_path = Path(cmd[file_arg_index])
            backup_path.parent.mkdir(parents=True, exist_ok=True)
            backup_path.write_text("mock backup content")
            return mock_result

        mock_run.side_effect = create_backup_file

        # Test backup creation with custom suffix
        custom_suffix = "pre_migration"
        backup_path = backup_tool.create_backup(custom_suffix)

        assert backup_path.exists()
        assert custom_suffix in backup_path.name
        assert backup_path.name.startswith("postgres_backup_")

    @patch("subprocess.run")
    def test_create_backup_failure_cleanup(
        self, mock_run: Mock, backup_tool: PostgreSQLBackupTool
    ):
        """Test backup creation failure triggers cleanup of partial files."""
        # Setup mock subprocess to fail
        mock_run.side_effect = Exception("pg_dump failed")

        # Test that backup creation raises exception
        with pytest.raises(RuntimeError, match="Backup creation failed"):
            backup_tool.create_backup()

        # Verify no partial backup files remain
        assert len(list(backup_tool.backup_dir.glob("*.dump"))) == 0
        assert len(list(backup_tool.backup_dir.glob("*.sql"))) == 0

    def test_backup_integrity_validation_success(
        self, backup_tool: PostgreSQLBackupTool, temp_dir: Path
    ):
        """Test backup integrity validation passes for valid backup."""
        # Create a mock backup file with content
        backup_path = temp_dir / "test_backup.dump"
        backup_path.write_text("valid backup content")

        # Create checksum file (required for validation)
        checksum = backup_tool.validator.generate_checksum(backup_path)
        backup_tool.validator.save_checksum(backup_path, checksum)

        # Create corresponding metadata file using validator
        metadata = backup_tool.validator.create_backup_metadata(
            backup_path=backup_path, database_version="14.0", backup_type="postgresql"
        )
        backup_tool.validator.save_metadata(backup_path, metadata)

        # Test validation
        is_valid = backup_tool.validate_backup_integrity(backup_path)
        assert is_valid is True

    def test_backup_integrity_validation_failure_missing_file(
        self, backup_tool: PostgreSQLBackupTool, temp_dir: Path
    ):
        """Test backup integrity validation fails for missing backup file."""
        backup_path = temp_dir / "nonexistent_backup.dump"

        is_valid = backup_tool.validate_backup_integrity(backup_path)
        assert is_valid is False

    def test_backup_integrity_validation_failure_corrupted_checksum(
        self, backup_tool: PostgreSQLBackupTool, temp_dir: Path
    ):
        """Test backup integrity validation fails for corrupted backup."""
        # Create a backup file
        backup_path = temp_dir / "test_backup.dump"
        backup_path.write_text("original backup content")

        # Create checksum for original content
        original_checksum = backup_tool.validator.generate_checksum(backup_path)
        backup_tool.validator.save_checksum(backup_path, original_checksum)

        # Create metadata with original checksum
        metadata = backup_tool.validator.create_backup_metadata(
            backup_path=backup_path, database_version="14.0", backup_type="postgresql"
        )
        backup_tool.validator.save_metadata(backup_path, metadata)

        # Corrupt the backup file (this will make checksum validation fail)
        backup_path.write_text("corrupted backup content")

        # Test validation fails
        is_valid = backup_tool.validate_backup_integrity(backup_path)
        assert is_valid is False

    @patch("subprocess.run")
    def test_restore_from_backup_success(
        self, mock_run: Mock, backup_tool: PostgreSQLBackupTool, temp_dir: Path
    ):
        """Test successful restoration from backup."""
        # Setup mock subprocess response
        mock_result = Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        mock_run.return_value = mock_result

        # Create a mock backup file with proper validation files
        backup_path = temp_dir / "test_backup.dump"
        backup_path.write_text("backup content")

        # Create checksum file
        checksum = backup_tool.validator.generate_checksum(backup_path)
        backup_tool.validator.save_checksum(backup_path, checksum)

        # Create metadata file
        metadata = backup_tool.validator.create_backup_metadata(
            backup_path=backup_path, database_version="14.0", backup_type="postgresql"
        )
        backup_tool.validator.save_metadata(backup_path, metadata)

        # Test restore operation
        backup_tool.restore_from_backup(backup_path)

        # Verify pg_restore was called
        mock_run.assert_called_once()
        cmd = mock_run.call_args[0][0]
        assert "pg_restore" in cmd or "psql" in cmd
        assert str(backup_path) in cmd

    @patch("subprocess.run")
    def test_restore_from_backup_to_different_database(
        self, mock_run: Mock, backup_tool: PostgreSQLBackupTool, temp_dir: Path
    ):
        """Test restore to different target database."""
        # Setup mock subprocess response
        mock_result = Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        mock_run.return_value = mock_result

        # Create a mock backup file with proper validation files
        backup_path = temp_dir / "test_backup.dump"
        backup_path.write_text("backup content")

        # Create checksum file
        checksum = backup_tool.validator.generate_checksum(backup_path)
        backup_tool.validator.save_checksum(backup_path, checksum)

        # Create metadata file
        metadata = backup_tool.validator.create_backup_metadata(
            backup_path=backup_path, database_version="14.0", backup_type="postgresql"
        )
        backup_tool.validator.save_metadata(backup_path, metadata)

        # Test restore to different database
        target_db = "different_db"
        backup_tool.restore_from_backup(backup_path, target_db)

        # Verify restore command included target database
        mock_run.assert_called_once()
        cmd = mock_run.call_args[0][0]
        assert target_db in cmd

    def test_list_available_backups_empty_directory(
        self, backup_tool: PostgreSQLBackupTool
    ):
        """Test listing backups in empty directory."""
        backups = backup_tool.list_available_backups()
        assert len(backups) == 0

    def test_list_available_backups_with_backups(
        self, backup_tool: PostgreSQLBackupTool
    ):
        """Test listing and sorting of available backups."""
        # Create multiple mock backup files with metadata
        backup_files = []
        for i in range(3):
            backup_path = (
                backup_tool.backup_dir / f"postgres_backup_2024010{i + 1}_120000.dump"
            )
            backup_path.write_text(f"backup content {i}")

            # Create metadata
            metadata = backup_tool.validator.create_backup_metadata(
                backup_path=backup_path,
                database_version="14.0",
                backup_type="postgresql",
            )
            # Override the timestamp for testing
            metadata["timestamp"] = datetime(
                2024, 1, i + 1, 12, 0, 0, tzinfo=timezone.utc
            ).isoformat()
            backup_tool.validator.save_metadata(backup_path, metadata)
            backup_files.append(backup_path)

        # List backups
        backups = backup_tool.list_available_backups()

        assert len(backups) == 3
        # Should be sorted by creation time (newest first)
        # Note: BackupInfo uses created_at field based on metadata timestamp
        backup_dates = [backup.created_at for backup in backups]
        assert len(set(backup_dates)) == 3  # All dates should be different

    def test_backup_rotation_policy(self, backup_tool: PostgreSQLBackupTool):
        """Test backup rotation removes oldest backups."""
        # Set max_backups to 2 for testing
        backup_tool.settings.max_backups = 2

        # Create 4 backup files
        backup_files = []
        for i in range(4):
            backup_path = (
                backup_tool.backup_dir / f"postgres_backup_2024010{i + 1}_120000.dump"
            )
            backup_path.write_text(f"backup content {i}")

            # Create metadata with different timestamps
            metadata = backup_tool.validator.create_backup_metadata(
                backup_path=backup_path,
                database_version="14.0",
                backup_type="postgresql",
            )
            # Override timestamp for testing
            metadata["timestamp"] = datetime(
                2024, 1, i + 1, 12, 0, 0, tzinfo=timezone.utc
            ).isoformat()
            backup_tool.validator.save_metadata(backup_path, metadata)
            backup_files.append(backup_path)

        # Trigger rotation
        backup_tool.manager.rotate_backups(
            backup_tool.backup_dir, backup_tool.settings.max_backups
        )

        # Only the 2 newest backups should remain
        remaining_backups = list(backup_tool.backup_dir.glob("*.dump"))
        assert len(remaining_backups) == 2

        # Verify the newest ones are kept
        remaining_names = {f.name for f in remaining_backups}
        assert "postgres_backup_20240103_120000.dump" in remaining_names
        assert "postgres_backup_20240104_120000.dump" in remaining_names

    def test_delete_backup_success(self, backup_tool: PostgreSQLBackupTool):
        """Test successful backup deletion."""
        # Create a backup file with metadata
        backup_path = backup_tool.backup_dir / "test_backup.dump"
        backup_path.write_text("backup content")

        metadata = backup_tool.validator.create_backup_metadata(
            backup_path=backup_path, database_version="14.0", backup_type="postgresql"
        )
        metadata_path = backup_tool.validator.save_metadata(backup_path, metadata)

        # Verify files exist
        assert backup_path.exists()
        assert metadata_path.exists()

        # Delete backup
        backup_tool.delete_backup(backup_path)

        # Verify files are deleted
        assert not backup_path.exists()
        assert not metadata_path.exists()

    def test_delete_backup_nonexistent_file(
        self, backup_tool: PostgreSQLBackupTool, temp_dir: Path
    ):
        """Test deletion of nonexistent backup file."""
        backup_path = temp_dir / "nonexistent_backup.dump"

        # Should not raise exception for nonexistent file
        backup_tool.delete_backup(backup_path)

    def test_concurrent_backup_operations_locking(
        self, backup_tool: PostgreSQLBackupTool
    ):
        """Test handling of concurrent backup operations with locking."""
        # This test would require more complex setup to actually test concurrency
        # For now, we'll test that the locking mechanism is properly initialized
        assert hasattr(backup_tool, "_lock_manager")

    @patch("subprocess.run")
    def test_backup_with_connection_failure(
        self, mock_run: Mock, backup_tool: PostgreSQLBackupTool
    ):
        """Test backup behavior when database connection fails."""
        # Setup mock to simulate connection failure
        mock_run.side_effect = Exception("connection failed")

        # Test that backup creation raises appropriate exception
        with pytest.raises(RuntimeError, match="Backup creation failed"):
            backup_tool.create_backup()

    def test_backup_formats_support(self, backup_settings: PostgreSQLBackupSettings):
        """Test different backup format support."""
        # Test custom format
        backup_settings.backup_format = "custom"
        tool_custom = PostgreSQLBackupTool(backup_settings)
        assert tool_custom.settings.backup_format == "custom"

        # Test plain format
        backup_settings.backup_format = "plain"
        tool_plain = PostgreSQLBackupTool(backup_settings)
        assert tool_plain.settings.backup_format == "plain"

        # Test tar format
        backup_settings.backup_format = "tar"
        tool_tar = PostgreSQLBackupTool(backup_settings)
        assert tool_tar.settings.backup_format == "tar"

    def test_backup_metadata_creation(self, backup_tool: PostgreSQLBackupTool):
        """Test backup metadata is created with correct structure."""
        # Create a mock backup file
        backup_path = backup_tool.backup_dir / "test_backup.dump"
        backup_path.write_text("backup content")

        # Create metadata
        backup_tool._create_backup_metadata(backup_path)

        # Verify metadata file exists and has correct structure
        metadata_path = backup_path.with_suffix(".metadata.json")
        assert metadata_path.exists()

        metadata = backup_tool.validator.load_metadata(backup_path)
        assert metadata is not None, "Metadata should not be None"
        assert metadata["backup_path"] == str(backup_path)
        assert metadata["backup_type"] == "postgresql"
        assert "timestamp" in metadata
        assert "checksum" in metadata
        assert "backup_size" in metadata

    def test_pg_dump_command_construction(self, backup_tool: PostgreSQLBackupTool):
        """Test pg_dump command is constructed correctly."""
        backup_path = backup_tool.backup_dir / "test_backup.dump"

        # Test command construction for different formats
        cmd_custom = backup_tool._build_pg_dump_command(backup_path, "custom")
        assert "pg_dump" in cmd_custom
        assert "--format=custom" in cmd_custom
        assert str(backup_path) in cmd_custom

        cmd_plain = backup_tool._build_pg_dump_command(backup_path, "plain")
        assert "--format=plain" in cmd_plain

    @patch.dict(os.environ, {}, clear=True)
    def test_environment_variable_isolation(self, backup_tool: PostgreSQLBackupTool):
        """Test that backup operations don't leak environment variables."""
        # Ensure clean environment
        assert "PGPASSWORD" not in os.environ

        # The backup tool should handle password securely without leaking
        # This is more of a structural test
        assert hasattr(backup_tool, "_parse_db_url")



"""
E2E Test Runner for MADSci.

Executes test definitions and captures results.

Trust boundary: test YAML files are trusted developer input (like Makefiles
or CI configs).  ``skip_if`` expressions are evaluated with ``simpleeval``
for defence-in-depth.  ``_execute_python()`` uses ``exec()`` and
``_execute_foreground_command()``/``_execute_background_command()`` use
``shell=True`` intentionally — they run commands authored by the test
developer, analogous to ``make`` recipe lines.
"""

import contextlib
import io
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
from datetime import datetime
from pathlib import Path
from typing import Any

import simpleeval
from madsci.common.testing.types import (
    E2ETestDefinition,
    E2ETestResult,
    E2ETestStep,
    StepResult,
    TestMode,
    ValidationConfig,
)
from madsci.common.testing.validators import (
    get_validator_registry,
    wait_for_condition,
)
from rich.console import Console


class E2ETestRunner:
    """
    Executes E2E test definitions.

    Supports both pure Python mode (no Docker) and Docker mode.
    """

    def __init__(
        self,
        working_dir: Path | None = None,
        mode: TestMode | None = None,
        console: Console | None = None,
        verbose: bool = False,
        capture_logs: bool = True,
    ) -> None:
        """
        Initialize the test runner.

        Args:
            working_dir: Base directory for test execution. If None, uses a temp dir.
            mode: Test execution mode (PYTHON, DOCKER, HYBRID). If None, uses test's mode.
            console: Rich console for output. If None, creates one.
            verbose: If True, print verbose output.
            capture_logs: If True, capture all logs to a file.
        """
        self.working_dir = working_dir or Path(tempfile.mkdtemp(prefix="madsci_e2e_"))
        self.mode_override = mode
        self.console = console or Console()
        self.verbose = verbose
        self.capture_logs = capture_logs
        self.validator_registry = get_validator_registry()
        self._background_processes: list[subprocess.Popen] = []
        self._logs_file: Path | None = None

    def run(
        self,
        test_definition: E2ETestDefinition,
        env: dict[str, str] | None = None,
    ) -> E2ETestResult:
        """
        Run an E2E test.

        Args:
            test_definition: The test to run
            env: Additional environment variables

        Returns:
            E2ETestResult with the test outcome
        """
        mode = self.mode_override or test_definition.mode
        started_at = datetime.now()

        # Prepare environment and logs
        test_env = self._prepare_environment(env)
        self._setup_logs(test_definition.name)
        self._print_test_header(test_definition.name, mode)

        # Execute test phases
        setup_results, step_results, passed, error = self._execute_test_phases(
            test_definition, test_env
        )

        # Cleanup and finalize
        cleanup_executed = self._run_cleanup(test_definition, test_env)
        completed_at = datetime.now()

        result = E2ETestResult(
            test_name=test_definition.name,
            passed=passed,
            mode=mode,
            started_at=started_at,
            completed_at=completed_at,
            duration_seconds=(completed_at - started_at).total_seconds(),
            setup_results=setup_results,
            step_results=step_results,
            cleanup_executed=cleanup_executed,
            error=error,
            logs_path=self._logs_file,
        )

        self._print_summary(result)
        return result

    def _prepare_environment(self, env: dict[str, str] | None) -> dict[str, str]:
        """Prepare the environment variables for test execution."""
        test_env = os.environ.copy()
        if env:
            test_env.update(env)
        return test_env

    def _setup_logs(self, test_name: str) -> None:
        """Set up log file for capturing output."""
        if self.capture_logs:
            # Ensure working directory exists
            self.working_dir.mkdir(parents=True, exist_ok=True)
            self._logs_file = (
                self.working_dir / f"{test_name.replace(' ', '_')}_logs.txt"
            )
            self._logs_file.touch()

    def _print_test_header(self, test_name: str, mode: TestMode) -> None:
        """Print the test header information."""
        self.console.print(f"\n[bold blue]Running test: {test_name}[/bold blue]")
        self.console.print(f"Mode: {mode.value}")
        self.console.print(f"Working directory: {self.working_dir}")

    def _execute_test_phases(
        self,
        test_definition: E2ETestDefinition,
        test_env: dict[str, str],
    ) -> tuple[list[StepResult], list[StepResult], bool, str | None]:
        """Execute all test phases (requirements check, setup, main steps)."""
        mode = self.mode_override or test_definition.mode
        setup_results: list[StepResult] = []
        step_results: list[StepResult] = []
        error: str | None = None
        passed = True

        try:
            # Check requirements
            if not self._check_requirements(test_definition, mode):
                return setup_results, step_results, False, "Requirements not met"

            # Run setup steps
            setup_results, passed, error = self._run_setup_steps(
                test_definition.setup, test_env
            )
            if not passed:
                return setup_results, step_results, passed, error

            # Run main test steps
            step_results, passed, error = self._run_main_steps(
                test_definition.steps, test_env
            )

        except KeyboardInterrupt:
            error = "Test interrupted by user"
            passed = False
        except Exception as e:
            error = f"Unexpected error: {e}"
            passed = False

        return setup_results, step_results, passed, error

    def _run_setup_steps(
        self,
        setup_steps: list[E2ETestStep],
        test_env: dict[str, str],
    ) -> tuple[list[StepResult], bool, str | None]:
        """Run setup steps and return results."""
        results: list[StepResult] = []
        if not setup_steps:
            return results, True, None

        self.console.print("\n[bold]Running setup steps...[/bold]")
        for step in setup_steps:
            result = self._run_step(step, test_env)
            results.append(result)
            self._print_step_result(result)
            if not result.passed and not step.continue_on_error:
                return results, False, f"Setup step failed: {step.name}"

        return results, True, None

    def _run_main_steps(
        self,
        steps: list[E2ETestStep],
        test_env: dict[str, str],
    ) -> tuple[list[StepResult], bool, str | None]:
        """Run main test steps and return results."""
        results: list[StepResult] = []
        self.console.print("\n[bold]Running test steps...[/bold]")
        any_failed = False

        for step in steps:
            result = self._run_step(step, test_env)
            results.append(result)
            self._print_step_result(result)
            if not result.passed:
                any_failed = True
                if not step.continue_on_error:
                    return results, False, f"Step failed: {step.name}"

        # If any step failed (even with continue_on_error), the test fails
        if any_failed:
            return results, False, None

        return results, True, None

    def run_from_yaml(self, yaml_path: Path, **kwargs: Any) -> E2ETestResult:
        """
        Load and run a test from a YAML file.

        Args:
            yaml_path: Path to the test definition YAML file
            **kwargs: Additional arguments passed to run()

        Returns:
            E2ETestResult with the test outcome
        """
        test_definition = E2ETestDefinition.from_yaml(yaml_path)
        return self.run(test_definition, **kwargs)

    def _check_requirements(
        self,
        test_definition: E2ETestDefinition,
        mode: TestMode,
    ) -> bool:
        """Check if all requirements are met for running the test."""
        reqs = test_definition.requirements

        # Check Docker requirement
        if (reqs.docker or mode == TestMode.DOCKER) and not self._is_docker_available():
            self.console.print("[red]Docker is required but not available[/red]")
            return False

        # Check Docker Compose requirement
        if reqs.docker_compose and not self._is_docker_compose_available():
            self.console.print(
                "[red]Docker Compose is required but not available[/red]"
            )
            return False

        # Check Python version (basic check)
        if reqs.python:
            current = f"{sys.version_info.major}.{sys.version_info.minor}"
            # Simple version check - could be more sophisticated
            if reqs.python.startswith(">="):
                required = reqs.python[2:]
                if current < required:
                    self.console.print(
                        f"[red]Python {reqs.python} required, but {current} found[/red]"
                    )
                    return False

        return True

    def _is_docker_available(self) -> bool:
        """Check if Docker is available."""
        try:
            result = subprocess.run(
                ["docker", "info"],  # noqa: S607
                check=False,
                capture_output=True,
                timeout=10,
            )
            return result.returncode == 0
        except (FileNotFoundError, subprocess.TimeoutExpired):
            return False

    def _is_docker_compose_available(self) -> bool:
        """Check if Docker Compose is available."""
        try:
            result = subprocess.run(
                ["docker", "compose", "version"],  # noqa: S607
                check=False,
                capture_output=True,
                timeout=10,
            )
            return result.returncode == 0
        except (FileNotFoundError, subprocess.TimeoutExpired):
            return False

    def _run_step(
        self,
        step: E2ETestStep,
        env: dict[str, str],
    ) -> StepResult:
        """Execute a single test step."""
        started_at = datetime.now()

        # Check skip condition
        skip_result = self._check_skip_condition(step, env, started_at)
        if skip_result is not None:
            return skip_result

        # Prepare working directory and environment
        working_dir = self._prepare_step_working_dir(step)
        step_env = self._prepare_step_env(env, step.env)

        # Execute command or Python code
        exit_code, stdout, stderr, error_msg = self._execute_step_action(
            step, working_dir, step_env
        )

        # Run validations and build result
        return self._build_step_result(
            step, started_at, working_dir, exit_code, stdout, stderr, error_msg
        )

    def _check_skip_condition(
        self,
        step: E2ETestStep,
        env: dict[str, str],
        started_at: datetime,
    ) -> StepResult | None:
        """Check if step should be skipped. Returns StepResult if skipped, None otherwise."""
        if not step.skip_if:
            return None

        try:
            evaluator = simpleeval.EvalWithCompoundTypes(
                names={
                    "env": env,
                    "os_name": os.name,
                    "os_sep": os.sep,
                    "platform": sys.platform,
                },
                functions={"get": lambda d, k, default=None: d.get(k, default)},
            )
            should_skip = evaluator.eval(step.skip_if)
            if should_skip:
                return StepResult(
                    step_name=step.name,
                    passed=True,
                    skipped=True,
                    skip_reason=f"Skip condition met: {step.skip_if}",
                    duration_seconds=0,
                    started_at=started_at,
                    completed_at=datetime.now(),
                )
        except Exception as e:
            return StepResult(
                step_name=step.name,
                passed=False,
                duration_seconds=0,
                started_at=started_at,
                completed_at=datetime.now(),
                error=f"Error evaluating skip condition: {e}",
            )
        return None

    def _prepare_step_working_dir(self, step: E2ETestStep) -> Path:
        """Prepare and return the working directory for a step."""
        working_dir = self.working_dir
        if step.working_dir:
            working_dir = self.working_dir / step.working_dir
            working_dir.mkdir(parents=True, exist_ok=True)
        return working_dir

    def _prepare_step_env(
        self, base_env: dict[str, str], step_env: dict[str, str]
    ) -> dict[str, str]:
        """Merge base environment with step-specific environment."""
        merged = base_env.copy()
        merged.update(step_env)
        return merged

    def _execute_step_action(
        self,
        step: E2ETestStep,
        working_dir: Path,
        step_env: dict[str, str],
    ) -> tuple[int | None, str, str, str | None]:
        """Execute the step's command or Python code."""
        if step.command:
            exit_code, stdout, stderr, error_msg = self._execute_command(
                step.command,
                working_dir,
                step_env,
                step.timeout,
                step.background,
            )

            # Wait for condition if background process
            if step.background and step.wait_for and exit_code == 0:
                wait_config = ValidationConfig(
                    type=step.wait_for.type,
                    url=step.wait_for.url,
                    path=step.wait_for.path,
                    timeout=step.wait_for.timeout,
                )
                context = {"working_dir": str(working_dir)}
                if not wait_for_condition(
                    wait_config,
                    context,
                    timeout=step.wait_for.timeout,
                    poll_interval=step.wait_for.poll_interval,
                ):
                    error_msg = (
                        f"Timeout waiting for condition: {step.wait_for.type.value}"
                    )

            return exit_code, stdout, stderr, error_msg

        if step.python_code:
            return self._execute_python(step.python_code, working_dir, step_env)

        return None, "", "", None

    def _build_step_result(
        self,
        step: E2ETestStep,
        started_at: datetime,
        working_dir: Path,
        exit_code: int | None,
        stdout: str,
        stderr: str,
        error_msg: str | None,
    ) -> StepResult:
        """Build and return the step result with validations."""
        completed_at = datetime.now()
        duration = (completed_at - started_at).total_seconds()

        # Run validations
        context = {
            "exit_code": exit_code,
            "stdout": stdout,
            "stderr": stderr,
            "working_dir": str(working_dir),
        }
        validation_results = [
            self.validator_registry.validate(validation, context)
            for validation in step.validations
        ]

        # Determine overall pass/fail
        passed = error_msg is None and all(v.passed for v in validation_results)

        # Log output if capturing
        self._log_step_output(step.name, stdout, stderr)

        return StepResult(
            step_name=step.name,
            passed=passed,
            exit_code=exit_code,
            stdout=stdout,
            stderr=stderr,
            duration_seconds=duration,
            started_at=started_at,
            completed_at=completed_at,
            validation_results=validation_results,
            error=error_msg,
        )

    def _log_step_output(self, step_name: str, stdout: str, stderr: str) -> None:
        """Log step output to the logs file."""
        if self._logs_file and (stdout or stderr):
            with self._logs_file.open("a") as f:
                f.write(f"\n=== Step: {step_name} ===\n")
                if stdout:
                    f.write(f"STDOUT:\n{stdout}\n")
                if stderr:
                    f.write(f"STDERR:\n{stderr}\n")

    def _execute_command(
        self,
        command: str,
        working_dir: Path,
        env: dict[str, str],
        timeout: float,
        background: bool = False,
    ) -> tuple[int | None, str, str, str | None]:
        """
        Execute a shell command.

        Returns:
            Tuple of (exit_code, stdout, stderr, error_message)
        """
        if self.verbose:
            self.console.print(f"  [dim]$ {command}[/dim]")

        try:
            if background:
                return self._execute_background_command(command, working_dir, env)
            return self._execute_foreground_command(command, working_dir, env, timeout)
        except subprocess.TimeoutExpired:
            return None, "", "", f"Command timed out after {timeout}s"
        except Exception as e:
            return None, "", "", f"Command execution error: {e}"

    def _execute_background_command(
        self,
        command: str,
        working_dir: Path,
        env: dict[str, str],
    ) -> tuple[int | None, str, str, str | None]:
        """Execute a command in the background.

        Uses ``shell=True`` intentionally — test YAML files are trusted input.
        """
        process = subprocess.Popen(  # noqa: S602
            command,
            shell=True,  # Required for running user-provided shell commands
            cwd=working_dir,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            start_new_session=True,
        )
        self._background_processes.append(process)
        # Give it a moment to start
        time.sleep(0.5)
        # Check if it crashed immediately
        poll = process.poll()
        if poll is not None and poll != 0:
            stdout, stderr = process.communicate()
            return (
                poll,
                stdout.decode(),
                stderr.decode(),
                "Background process exited immediately",
            )
        return 0, "", "", None

    def _execute_foreground_command(
        self,
        command: str,
        working_dir: Path,
        env: dict[str, str],
        timeout: float,
    ) -> tuple[int, str, str, str | None]:
        """Execute a command in the foreground.

        Uses ``shell=True`` intentionally — test YAML files are trusted input.
        """
        result = subprocess.run(  # noqa: S602
            command,
            check=False,
            shell=True,  # Required for running user-provided shell commands
            cwd=working_dir,
            env=env,
            capture_output=True,
            text=True,
            timeout=timeout,
        )
        return result.returncode, result.stdout, result.stderr, None

    def _execute_python(
        self,
        code: str,
        working_dir: Path,
        env: dict[str, str],
    ) -> tuple[int, str, str, str | None]:
        """
        Execute Python code.

        Uses ``exec()`` intentionally — test YAML files are trusted input.

        Returns:
            Tuple of (exit_code, stdout, stderr, error_message)
        """
        old_cwd = Path.cwd()
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        stdout_capture: io.StringIO = io.StringIO()
        stderr_capture: io.StringIO = io.StringIO()

        try:
            os.chdir(working_dir)
            sys.stdout = stdout_capture
            sys.stderr = stderr_capture

            exec(code, {"env": env, "Path": Path, "os": os})  # noqa: S102

            stdout = stdout_capture.getvalue()
            stderr = stderr_capture.getvalue()
            return 0, stdout, stderr, None
        except Exception as e:
            stderr = stderr_capture.getvalue()
            return 1, "", stderr, str(e)
        finally:
            os.chdir(old_cwd)
            sys.stdout = old_stdout
            sys.stderr = old_stderr

    def _run_cleanup(
        self,
        test_definition: E2ETestDefinition,
        env: dict[str, str],
    ) -> bool:
        """Run cleanup actions."""
        cleanup = test_definition.cleanup
        executed = False

        # Stop background processes
        if cleanup.stop_background:
            self._stop_background_processes()
            executed = True

        # Run cleanup commands
        if self._run_cleanup_commands(cleanup.commands, env):
            executed = True

        # Remove files
        if self._cleanup_files(cleanup.files):
            executed = True

        return executed

    def _stop_background_processes(self) -> None:
        """Stop all background processes."""
        for process in self._background_processes:
            try:
                os.killpg(os.getpgid(process.pid), signal.SIGTERM)
                process.wait(timeout=5)
            except Exception:
                with contextlib.suppress(Exception):
                    os.killpg(os.getpgid(process.pid), signal.SIGKILL)
        self._background_processes.clear()

    def _run_cleanup_commands(self, commands: list[str], env: dict[str, str]) -> bool:
        """Run cleanup commands. Returns True if any were executed."""
        executed = False
        for command in commands:
            try:
                subprocess.run(  # noqa: S602
                    command,
                    check=False,
                    shell=True,  # Required for running user-provided shell commands
                    cwd=self.working_dir,
                    env=env,
                    capture_output=True,
                    timeout=30,
                )
                executed = True
            except Exception:  # noqa: S110
                pass  # Best effort cleanup - intentionally ignoring exceptions
        return executed

    def _cleanup_files(self, file_patterns: list[str]) -> bool:
        """Remove files matching patterns. Returns True if any were processed."""
        executed = False
        for file_pattern in file_patterns:
            try:
                for file_path in self.working_dir.glob(file_pattern):
                    if file_path.is_file():
                        file_path.unlink()
                    elif file_path.is_dir():
                        shutil.rmtree(file_path)
                executed = True
            except Exception:  # noqa: S110
                pass  # Best effort cleanup - intentionally ignoring exceptions
        return executed

    def _print_step_result(self, result: StepResult) -> None:
        """Print the result of a step."""
        if result.skipped:
            self.console.print(f"  [yellow]⊘ {result.step_name}[/yellow] (skipped)")
            return

        status = "[green]✓[/green]" if result.passed else "[red]✗[/red]"
        self.console.print(
            f"  {status} {result.step_name} ({result.duration_seconds:.2f}s)"
        )

        if not result.passed:
            if result.error:
                self.console.print(f"    [red]Error: {result.error}[/red]")
            for validation in result.validation_results:
                if not validation.passed:
                    self.console.print(f"    [red]- {validation.message}[/red]")
            if result.stderr:
                self.console.print(f"    [dim]STDERR:[/dim]\n{result.stderr.strip()}")
            if result.stdout:
                self.console.print(f"    [dim]STDOUT:[/dim]\n{result.stdout.strip()}")

        if self.verbose:
            for validation in result.validation_results:
                status = "[green]✓[/green]" if validation.passed else "[red]✗[/red]"
                self.console.print(f"    {status} {validation.message}")

    def _print_summary(self, result: E2ETestResult) -> None:
        """Print the test summary."""
        self.console.print("\n" + "=" * 60)
        status = (
            "[bold green]PASSED[/bold green]"
            if result.passed
            else "[bold red]FAILED[/bold red]"
        )
        self.console.print(f"Test: {result.test_name} - {status}")
        self.console.print(f"Duration: {result.duration_seconds:.2f}s")
        self.console.print(
            f"Steps: {result.steps_passed} passed, {result.steps_failed} failed, "
            f"{result.steps_skipped} skipped"
        )
        if result.error:
            self.console.print(f"[red]Error: {result.error}[/red]")
        if result.logs_path:
            self.console.print(f"Logs: {result.logs_path}")
        self.console.print("=" * 60 + "\n")
